/*   File Header   ---------------------------------
     File Name:        scene_intf.c
     Author and Date:  Glenn Heinle
     Description:      Client interface routines for the Scene 
                       Generator and Scene data structure
     Functions:
      GB_id  GI_create_display ()
      GB_id  GI_create_X_display ()
      void   GI_get_display_size ()
     
     End Header   ---------------------------------- */


#include <geodata.h>
#include <geodata_private.h>
#include <stdio.h>
#include <types.h>

extern int  GS_allocate_display ();
extern int  GS_generate_scene_display_wrapper_from_existing_x_window ();
extern void GS_display_size ();


/*   Function Header   -----------------------------
     Name:             GI_create_display
     Description:      Allocate and return a display. 
     Parameters:       name:   character string Name for the display
                       type:   Type of display to create.
		       width:  integer Width of an X window display.
		       height: integer Height of an X window display.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   A new display is allocated.  If the type is GS_XWINDOWS
                       then a new window is created on the default X display.
     Return Value:     GB_id of the new display.
     Logic:
       1.  Check width and height greater than 0.
       2.  Check for valid display type, reporting errors.
       3.  Add the display to GB.
       4.  Fill in the structure elements.
       5.  Call the executor function.
     End Header   ---------------------------------- */

GB_id GI_create_display (label, name, type, width, height)
     char *label, *name;
     int type, width, height;
{
  GS_display_t new_display;
  GB_id id;
  int i;
  char message[100];

  if ((1 > width) || (1 > height)) {
    sprintf(message, "Width %d, or height %d, not greater than 0.",
	    width, height);
    GB_message (GB_ERROR, "GI_create_display", message);
  }
  else {
    switch (type) {
    case GS_CMFB:
    case GS_SMOOTH_CMFB:
    case GS_XWINDOWS:
    case GS_GLWINDOWS:
      new_display = (GS_display_t) GB_add_object (label, GB_DISPLAY);
      new_display->display_type = type;
      for (i=0; ((i<GD_FILE_NAME_LENGTH) || (name[i]==0)); i++)
	new_display->name[i] = name[i];
      new_display->size[0] = width;
      new_display->size[1] = height;
      GB_execute_display (new_display);
      id = GB_get_id (label, GB_DISPLAY);
      return (id);
      break;
    default: sprintf(message, "Invalid Display Type: %d", type);
      GB_message (GB_ERROR, "GI_create_display", message);
    }
  }
}


/*   Function Header   -----------------------------
     Name:             GI_get_display_id
     Description:      Locate a display with the given label.
     Parameters:       label: A character string containing the access label
                              for the display.  Must not be a null string.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   (None)
     Return Value:     The GB_id of the display with the given label if
                       it exists
     Logic:
       1.  Call GB_get_id with the GB_DISPLAY type.
     End Header   ---------------------------------- */

GB_id GI_get_display_id (label)
     char *label;
{
  return (GB_get_id (label, GB_DISPLAY));
}


/*   Function Header   -----------------------------
     Name:             GI_create_X_display
     Description:      Create a display out of an existing X window.
     Parameters:       name:   character string Name for the display
                       window: An existing X window.
     Preconditions:    Window is a valid X window.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   A new display is created out of an existing X window.
                       The given window will be used for GeoData output.
     Return Value:     GB_id of the new display.
     Logic:
       1.  Check for null window, reporting errors.
       2.  Call appropriate GS function.
     End Header   ---------------------------------- */

GB_id GI_create_X_display (name, window)
     char *name;
     Window window;
{
  if (window)
    return (GS_generate_scene_display_wrapper_from_existing_x_window
	    (name, window));
  else
    GB_message (GB_ERROR, "GI_create_display", "Cannot use a NULL window.");
}


/*   Function Header   -----------------------------
     Name:             GI_get_display_size
     Description:      Get the width and height of a display.
     Parameters:       display: GB_id of the display to get width and height.
                       width:   pointer to an integer used for return value.
		       height:  pointer to an integer used for return value.
     Preconditions:    display is a valid display.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   Width and height are set according to the width and
                       height of the given display.
     Return Value:     (None)
     Logic:
       1.  Check for jull display, reporting errors.
       2.  Call GS_display_size.
     End Header   ---------------------------------- */

void GI_get_display_size (display, width, height)
     GB_id display;
     int *width, *height;
{
  if (display)
    GS_display_size (display, width, height);
  else
    GB_message (GB_ERROR, "GI_create_display",
		"Cannot get size of a NULL display.");
}
