/*   File Header   ---------------------------------
     File Name:    ci/library.c
     Author and Date:  Paul B. Anderson/Glenn Heinle
                       6/90
     Description:  Various shared routines for client
                   interface routines
     End Header   ---------------------------------- */


/*  INCLUDES  */

#include <geodata.h>
#include <geodata_private.h>
#include <stdio.h>
#include <math.h>

double gmod(x,y)
  double x,y;
{
  double quotient, remainder;  int k;
  quotient = x/y;
  remainder = modf(quotient,&k)*y;
  return remainder;
}

/*  ====================  Library Routines  ====================  */

/*   Function Header   -----------------------------
     Description:      Retrieves and return a pointer to an object's
                       instance structure
     Parameters:       place:   A character string giving the calling
                                routine name, used for error messages
                       type:    An object type code, used for retrieval
                       id:      An integer object id, used for retrieval
     Preconditions:    The hash table must be set up
     Postconditions:   Error messages are generated if an object
                       with the supplied id does not exist or if it is
                       not the same type as the supplied type.  
     Return Value:     The function returns NULL if an object of the
                       desired type was not found.  Otherwise, an (int *)
                       pointer to the instance structure is returned.
     Logic:
     End Header   ---------------------------------- */

int *GI_get_instance_internal(place,type,id)
  char *place;  unsigned type;  GB_id id; 
{
  GB_entry *entry;  char etext[100];  int *result;

  result = NULL;
  entry = (GB_entry *)GB_get_entry(id);
  if (entry == NULL) {
    sprintf(etext,"ID is unknown: %d",id);
    GB_message(GB_ERROR,place,etext); }
  else {
    if(entry->type != type) {
      sprintf(etext,"ID %d, label %s, is type %d, not expected type %d",
              id,entry->label,entry->type,type);
      GB_message(GB_ERROR,place,etext); }
    else {
      result = entry->instance;
    }
  }

  return(result);
}


/*   Function Header   -----------------------------
     Description:      Allocate, a projection structure, format it with 
                       information for a "scaled geographics" projection,
                       and return a pointer to it
     Parameters:       (None)
     Preconditions:    None
     Postconditions:   A structure of type GI_proj_desc has been allocated
                       which must be deallocated by someone
     Return Value:     Pointer to a structure of type GI_proj_desc.
     Logic:            
       1.  Allocate a structure of type GI_proj_desc.  If malloc
           can't allocate space, return NULL and discontinue processing.
       2.  Fill in projection descriptor and return it.
     End Header   ---------------------------------- */

GI_proj_desc *GI_scaled_geographics()
{
  GI_proj_desc *result;  char etext[100];
  char *place = "GI_scaled_geographics";

  result = (GI_proj_desc *)malloc(sizeof(GI_proj_desc));

  if (result == NULL) {
    sprintf(etext,"Malloc couldn't allocate space");
    GB_message(GB_ERROR,place,etext); }
  else {
    result->code = GS_NO_PROJECTION;
    result->param[0] = 0.0;
    result->param[1] = 0.0;
    result->param[2] = 0.0;
    result->param[3] = 0.0;
    result->param[4] = 0.0;
    result->param[5] = 0.0;
    result->param[6] = 0.0;
    result->param[7] = 0.0;
    result->param[8] = 0.0;
    result->param[9] = 0.0;
  }

  return(result);
}


/*   Function Header   -----------------------------
     Description:      Allocate, a projection structure, format it with 
                       information for a Miller projection,
                       and return a pointer to it
     Parameters:       (None)
     Preconditions:    None
     Postconditions:   A structure of type GI_proj_desc has been allocated
                       which must be deallocated by someone
     Return Value:     Pointer to a structure of type GI_proj_desc.
     Logic:            
       1.  Allocate a structure of type GI_proj_desc.  If malloc
           can't allocate space, return NULL and discontinue processing.
       2.  Fill in projection descriptor and return it.
     End Header   ---------------------------------- */

GI_proj_desc *GI_miller()
{
  GI_proj_desc *result;  char etext[100];
  char *place = "GI_miller";

  result = (GI_proj_desc *)malloc(sizeof(GI_proj_desc));

  if (result == NULL) {
    sprintf(etext,"Malloc couldn't allocate space");
    GB_message(GB_ERROR,place,etext); }
  else {
    result->code = GS_MILLER;
    result->param[0] = 0.0;
    result->param[1] = 0.0;
    result->param[2] = 0.0;
    result->param[3] = 0.0;
    result->param[4] = 0.0;
    result->param[5] = 0.0;
    result->param[6] = 0.0;
    result->param[7] = 0.0;
    result->param[8] = 0.0;
    result->param[9] = 0.0;
  }

  return(result);
}

/*   Function Header   -----------------------------
     Description:      Allocate, a projection structure, format it with information
                       for a gnomonic projection, and return a pointer to it
     Parameters:       lat:      Latitude of point of tangency for
                                 projection plane (this must be in radians)
                       lon:      Longitude of point of tangency (radians)
     Preconditions:    None
     Postconditions:   A structure of type GI_proj_desc has been allocated
                       which must be deallocated by someone
     Return Value:     Pointer to a structure of type GI_proj_desc.
     Logic:            
       1.  Allocate projection structure.  If an error is an encountered,
           issue error message, return NULL, and discontinue processing.
       2.  Validate that the latitude is within legal range.  If
           not, issue error message and discontinue processing.
       3.  Fill in projection descriptor (mapping longitude
           into canonical range) and return it.
     End Header   ---------------------------------- */

GI_proj_desc *GI_gnomonic(lat,lon)
  float lat,lon;
{
  GI_proj_desc *result;  char etext[100];
  char *place = "GI_gnomonic";

  result = (GI_proj_desc *)malloc(sizeof(GI_proj_desc));

  if (result == NULL) {
    sprintf(etext,"Malloc couldn't allocate space");
    GB_message(GB_ERROR,place,etext); }
  else {
    if(lat < -0.5*GS_PI || lat > 0.5*GS_PI) {
      sprintf(etext,"Latitude is out of range: %f",lat);
      GB_message(GB_ERROR,place,etext); }
    else {
      result->code = GS_GNOMONIC;
      result->param[0] = lat;
      result->param[1] = gmod(lon,2.0*GS_PI);
      if (result->param[1] < -GS_PI) 
        result->param[1] += 2.0*GS_PI;
      if (result->param[1] > GS_PI) 
        result->param[1] -= 2.0*GS_PI;
      result->param[2] = 0.0;
      result->param[3] = 0.0;
      result->param[4] = 0.0;
      result->param[5] = 0.0;
      result->param[6] = 0.0;
      result->param[7] = 0.0;
      result->param[8] = 0.0;
      result->param[9] = 0.0;
    }
  }

  return(result);
}

/*   Function Header   -----------------------------
     Description:      Allocate, a projection structure, format it with information
                       for a mollweide projection, and return a pointer to it
     Parameters:       lon:      Central meridian (radians)
     Preconditions:    None
     Postconditions:   A structure of type GI_proj_desc has been allocated
                       which must be deallocated by someone
     Return Value:     Pointer to a structure of type GI_proj_desc.
     Logic:            
       1.  Allocate projection structure.  If an error is an encountered,
           issue error message, return NULL, and discontinue processing.
       2.  Fill in projection descriptor (mapping longitude
           into canonical range) and return it.
     End Header   ---------------------------------- */

GI_proj_desc *GI_mollweide(lon)
  float lon;
{
  GI_proj_desc *result;  char etext[100];
  char *place = "GI_mollweide";

  result = (GI_proj_desc *)malloc(sizeof(GI_proj_desc));

  if (result == NULL) {
    sprintf(etext,"Malloc couldn't allocate space");
    GB_message(GB_ERROR,place,etext); }
  else {
    result->code = GS_MOLLWEIDE;
    result->param[0] = gmod(lon,2.0*GS_PI);
    if (result->param[0] < -GS_PI) 
      result->param[0] += 2.0*GS_PI;
    if (result->param[0] > GS_PI) 
      result->param[0] -= 2.0*GS_PI;
    result->param[1] = 0.0;
    result->param[2] = 0.0;
    result->param[3] = 0.0;
    result->param[4] = 0.0;
    result->param[5] = 0.0;
    result->param[6] = 0.0;
    result->param[7] = 0.0;
    result->param[8] = 0.0;
    result->param[9] = 0.0;
  }

  return(result);
}


/*   Function Header   -----------------------------
     Description:      Allocate, a projection structure, format it with information
                       for a North Polar stereo projection, and return a pointer to it
     Parameters:       (None)
     Preconditions:    None
     Postconditions:   A structure of type GI_proj_desc has been allocated
                       which must be deallocated by someone
     Return Value:     Pointer to a structure of type GI_proj_desc.
     Logic:            
       1.  Allocate a structure of type GI_proj_desc.  If malloc
           can't allocate space, return NULL and discontinue processing.
       2.  Fill in projection descriptor and return it.
     End Header   ---------------------------------- */

GI_proj_desc *GI_np_stereo()
{
  GI_proj_desc *result;  char etext[100];
  char *place = "GI_np_stereo";

  result = (GI_proj_desc *)malloc(sizeof(GI_proj_desc));

  if (result == NULL) {
    sprintf(etext,"Malloc couldn't allocate space");
    GB_message(GB_ERROR,place,etext); }
  else {
    result->code = GS_POLAR_STEREO;
    result->param[0] = 1.0;
    result->param[1] = 0.0;
    result->param[2] = 0.0;
    result->param[3] = 0.0;
    result->param[4] = 0.0;
    result->param[5] = 0.0;
    result->param[6] = 0.0;
    result->param[7] = 0.0;
    result->param[8] = 0.0;
    result->param[9] = 0.0;
  }

  return(result);
}

/*   Function Header   -----------------------------
     Description:      Allocate, a projection structure, format it with information
                       for a South Polar stereo projection, and return a pointer to it
     Parameters:       (None)
     Preconditions:    None
     Postconditions:   A structure of type GI_proj_desc has been allocated
                       which must be deallocated by someone
     Return Value:     Pointer to a structure of type GI_proj_desc.
     Logic:            
       1.  Allocate a structure of type GI_proj_desc.  If malloc
           can't allocate space, return NULL and discontinue processing.
       2.  Fill in projection descriptor and return it.
     End Header   ---------------------------------- */

GI_proj_desc *GI_sp_stereo()
{
  GI_proj_desc *result;  char etext[100];
  char *place = "GI_sp_stereo";

  result = (GI_proj_desc *)malloc(sizeof(GI_proj_desc));

  if (result == NULL) {
    sprintf(etext,"Malloc couldn't allocate space");
    GB_message(GB_ERROR,place,etext); }
  else {
    result->code = GS_POLAR_STEREO;
    result->param[0] = -1.0;
    result->param[1] = 0.0;
    result->param[2] = 0.0;
    result->param[3] = 0.0;
    result->param[4] = 0.0;
    result->param[5] = 0.0;
    result->param[6] = 0.0;
    result->param[7] = 0.0;
    result->param[8] = 0.0;
    result->param[9] = 0.0;
  }

  return(result);
}


/*   Function Header   -----------------------------
     Description:      Allocate, a projection structure, 
                       format it with information for a 
		       Lambert conformal conic projection, 
		       and return a pointer to it.
     Parameters:       lat:      Latitude of point of tangency for
                                 projection plane (this must be in radians)
                       lon:      Longitude of point of tangency (radians)
     Preconditions:    None
     Postconditions:   A structure of type GI_proj_desc has been allocated
                       which must be deallocated by someone
     Return Value:     Pointer to a structure of type GI_proj_desc.
     Logic:            
       1.  Allocate projection structure.  If an error is an encountered,
           issue error message, return NULL, and discontinue processing.
       2.  Validate that all latitudes are within legal range.  If
           not, issue error message and discontinue processing.
       3.  Verify that south latitude is below north latitude.  If
           not, issue error message and discontinue processing.
       4.  Fill in projection descriptor (mapping longitude
           into canonical range) and return it.
     End Header   ---------------------------------- */

GI_proj_desc *GI_lambert(std_lat_1,std_lat_2,n_lat,s_lat,e_lon,w_lon)
  float std_lat_1,std_lat_2,n_lat,s_lat,e_lon,w_lon;
{
  GI_proj_desc *result;  char etext[100];
  char *place = "GI_lambert";

  result = (GI_proj_desc *)malloc(sizeof(GI_proj_desc));

  if (result == NULL) {
    sprintf(etext,"Malloc couldn't allocate space");
    GB_message(GB_ERROR,place,etext); }
  else {
    if(std_lat_1 < -0.5*GS_PI || std_lat_1 > 0.5*GS_PI) {
      sprintf(etext,"First standard latitude is out of range: %f",std_lat_1);
      GB_message(GB_ERROR,place,etext); }
    else {
      if(std_lat_2 < -0.5*GS_PI || std_lat_2 > 0.5*GS_PI) {
        sprintf(etext,"Second standard latitude is out of range: %f",std_lat_2);
        GB_message(GB_ERROR,place,etext); }
      else {
        if(n_lat < -0.5*GS_PI || n_lat > 0.5*GS_PI) {
          sprintf(etext,"North bounding latitude is out of range: %f",n_lat);
          GB_message(GB_ERROR,place,etext); }
        else {
          if(s_lat < -0.5*GS_PI || s_lat > 0.5*GS_PI) {
            sprintf(etext,"South bounding latitude is out of range: %f",s_lat);
            GB_message(GB_ERROR,place,etext); }
          else {
            if(s_lat >= n_lat) {
              sprintf(etext,"South latitude above North bound: %f, %f",
                      s_lat,n_lat);
              GB_message(GB_ERROR,place,etext); }
            else {
              result->code = GS_LAMBERT;
              result->param[0] = std_lat_1;
              result->param[1] = std_lat_2;
              result->param[2] = n_lat;
              result->param[3] = s_lat;
              result->param[4] = gmod(e_lon,2.0*GS_PI);
              if (result->param[4] < -GS_PI) 
                  (result->param[4]) += 2.0*GS_PI;
              if (result->param[4] > GS_PI) 
                result->param[4] -= 2.0*GS_PI;
              result->param[5] = gmod(w_lon,2.0*GS_PI);
              if (result->param[5] < -GS_PI) 
                result->param[5] += 2.0*GS_PI;
              if (result->param[5] > GS_PI) 
                result->param[5] -= 2.0*GS_PI;
              result->param[6] = 0.0;
              result->param[7] = 0.0;
              result->param[8] = 0.0;
              result->param[9] = 0.0;
            }
          }
        }
      }
    }
  }

  return(result);
}


/*   Function Header   -----------------------------
     Description:      Allocate, a projection structure, format it with information
                       for a perspective projection, and return a pointer to it
     Parameters:       lat:      Latitude of viewpoint (this must be in radians)
                       lon:      Longitude of viewpoint (radians)
                       alt:      Altitude of view point (in the units given by
                                 the units parameter below)
                       units:    Integer giving the altitude units
                       gamma:    View angle in plane orthogonal to nadir
                                 direction measured clockwise from North
                                 in radians
                       omega:    Pitch view angle measured from the nadir
                                 direction upward toward the plane of the
                                 gamma angle (radians)
     Preconditions:    None
     Postconditions:   A structure of type GI_proj_desc has been allocated
                       which must be deallocated by someone
     Return Value:     Pointer to a structure of type GI_proj_desc.
     Logic:            
       1.  Allocate projection structure.  If an error is an encountered,
           issue error message, return NULL, and discontinue processing.
       2.  Validate that latitude is within legal range.  If
           not, issue error message and discontinue processing.
       3.  Validate that units is within legal range.  If
           not, issue error message and discontinue processing.
       4.  Fill in projection descriptor, mapping longitude
           into canonical range, converting altitude to meters.
       5.  Return structure.
     End Header   ---------------------------------- */

GI_proj_desc *GI_globe_view(lat,lon,alt,units,gamma,omega)
  float lat,lon,alt,gamma,omega;  int units;
{
  char *place = "GI_globe_view";
  GI_proj_desc *result;  char etext[100];

  result = (GI_proj_desc *)malloc(sizeof(GI_proj_desc));

  if (result == NULL) {
    sprintf(etext,"Malloc couldn't allocate space");
    GB_message(GB_ERROR,place,etext); }
  else {
    if (lat < -0.5*GS_PI || lat > 0.5*GS_PI) {
      sprintf(etext,"Center latitude is out of range: %f",lat);
      GB_message(GB_ERROR,place,etext); }
    else {
      if (units != GD_METERS && units != GD_FEET &&
         units != GD_KM && units != GD_MILES && units != GD_EARTH_RADII) {
        sprintf(etext,"Altitude units choice is not valid: %d",units);
        GB_message(GB_ERROR,place,etext); }
      else {
        result->code = GS_GLOBE_VIEW;
        result->param[0] = lat;
        result->param[1] = gmod(lon,2.0*GS_PI);
        if (result->param[1] < -GS_PI) 
            result->param[1] += 2.0*GS_PI;
        if (result->param[1] > GS_PI) 
          result->param[1] -= 2.0*GS_PI;
        switch (units) {
        case GD_METERS:       result->param[2] = alt*1.0/GS_EARTH_RADIUS;
                              break;
        case GD_KM:           result->param[2] = alt*1000.0/GS_EARTH_RADIUS;
                              break;
        case GD_MILES:        result->param[2] = alt*6080.0*0.3048/GS_EARTH_RADIUS;
                              break;
        case GD_FEET:         result->param[2] = alt*0.3048/GS_EARTH_RADIUS;
                              break;
        case GD_EARTH_RADII:  result->param[2] = alt;
                              break;
        }
        result->param[3] = gamma;
        result->param[4] = omega;
        result->param[5] = 0.0;
        result->param[6] = 0.0;
        result->param[7] = 0.0;
        result->param[8] = 0.0;
        result->param[9] = 0.0;
      }
    }
  }

  return(result);
}


/* Function Header ----------------------------- 
	Description:		
		Allocate, a projection structure, format it with
		information for a perspective projection, and return a
		pointer to it
	Parameters: 	
		All coordinates are in Earth_Radii.
		x1,y1,z1: 	
			Cartesian coordinates of the "from" point
		x2,y2,z2:	
			Cartesian coordinates of the "to" point
		x_up,y_up,z_up:	
			Vector describing "up" ie top of screen.
		fov:		
			Field Of View in degrees.
		hither,yon:	
			Distance in Earth_Radii of front and rear
			clipping planes.
	Preconditions: 		
		None 
	Postconditions: 	
		A structure of type GI_proj_desc has been allocated
		which must be deallocated by someone
	Return Value: 		
		Pointer to a structure of type GI_proj_desc.
	Logic: 			
		1.  Allocate projection structure.  If an error is an
		    encountered, issue error message, return NULL.
		2.  Fill in parameters and set code.
		3.  Return structure.
End Header ---------------------------------- */

GI_proj_desc *GI_general_perspective(x1,y1,z1,x2,y2,z2,x_up,y_up,z_up,
	fov,hither,yon)
  float x1,y1,z1,x2,y2,z2,x_up,y_up,z_up,fov,hither,yon;
{
  char *place = "GI_general_perspective";
  GI_proj_desc *result;  char etext[100];

  result = (GI_proj_desc *)malloc(sizeof(GI_proj_desc));

  if (result == NULL) {
    sprintf(etext,"Malloc couldn't allocate space");
    GB_message(GB_ERROR,place,etext); }
  else {
    result->code = GS_GENERAL_PERSPECTIVE;
    result->param[0]=x1;
    result->param[1]=y1;
    result->param[2]=z1;
    result->param[3]=x2;
    result->param[4]=y2;
    result->param[5]=z2;
    result->param[6]=x_up;
    result->param[7]=y_up;
    result->param[8]=z_up;
    result->param[9]=fov;
    result->param[10]=hither;
    result->param[11]=yon;
  }
  return(result);
}

/*   Function Header   -----------------------------
     Description:      Allocate, a projection structure, 
                       format it with information for a null projection,
                       and return a pointer to it
     Parameters:       (None)
     Preconditions:    None
     Postconditions:   A structure of type GI_proj_desc has been allocated
                       which must be deallocated by someone
     Return Value:     Pointer to a structure of type GI_proj_desc.
     Logic:            
       1.  Allocate a structure of type GI_proj_desc.  If malloc
           can't allocate space, return NULL and discontinue processing.
       2.  Fill in projection descriptor and return it.
     End Header   ---------------------------------- */

GI_proj_desc *GI_none()
{
  GI_proj_desc *result;  char etext[100];
  char *place = "GI_none";

  result = (GI_proj_desc *)malloc(sizeof(GI_proj_desc));

  if (result == NULL) {
    sprintf(etext,"Malloc couldn't allocate space");
    GB_message(GB_ERROR,place,etext); }
  else {
    result->code = GS_NO_PROJECTION;
    result->param[0] = 0.0;
    result->param[1] = 0.0;
    result->param[2] = 0.0;
    result->param[3] = 0.0;
    result->param[4] = 0.0;
    result->param[5] = 0.0;
    result->param[6] = 0.0;
    result->param[7] = 0.0;
    result->param[8] = 0.0;
    result->param[9] = 0.0;
  }

  return(result);
}
