
/*   File Header   ---------------------------------
     File Name:        scene_intf.c
     Author and Date:  Glenn Heinle
     Description:      Client interface routines for the Scene 
                       Generator and Scene data structure
     Functions:
      GS_scene_t  GI_get_scene_internal ()
      GB_id       GI_get_scene_id ()
      GB_id       GI_create_scene ()
      void        GI_delete_scene ()
      void        GI_update_scene ()
      void        GI_display_scene ()
      int         GI_get_scene_projection ()
      void        GI_set_scene_projection ()
      void        GI_get_scene_projection_params ()
      void        GI_set_scene_projection_params ()
      float       GI_get_scene_projection_parameter ()
      void        GI_set_scene_projection_parameter ()
      void        GI_add_scene_data ()
      void        GI_delete_scene_data ()
      void        GI_resize_scene ()
      int         GI_get_scene_width ()
      void        GI_set_scene_width ()
      int         GI_get_scene_height ()
      void        GI_set_scene_height ()
      int         GI_get_scene_zoom ()
      void        GI_set_scene_zoom ()
      void        GI_get_scene_center ()
      void        GI_set_scene_center ()
      void        GI_get_scene_background ()
      void        GI_set_scene_background ()

     End Header   ---------------------------------- */


#include <geodata.h>
#include <geodata_private.h>
#include <types.h>

#include <stdio.h>

extern GS_representation_t GI_get_rep_internal ();
extern void GS_update_scene ();
extern void GS_display_scene ();


/*   Function Header   -----------------------------
     Name:             GI_get_scene_internal
     Description:      Return a pointer to a GS_scene structure.
     Parameters:       place:   A character string giving the calling
                                routine name, used for error messages.
		       id:      An integer object id, used for retrieval.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   (None)
     Return Value:     Return a pointer to a GS_scene structure.
                       NULL if a GS_scene was not found with the given id.
     Logic:
       1.  Calls GI_get_instance_internal and casts value into a GS_scene_t.
     End Header   ---------------------------------- */

GS_scene_t GI_get_scene_internal (place, id)
     char *place;
     GB_id id;
{
  return ((GS_scene_t)
	  GI_get_instance_internal (place, GB_SCENE, id));
}


/*   Function Header   -----------------------------
     Name:             GI_get_scene_id
     Description:      Locate a scene with the given label.
     Parameters:       label: A character string containing the access label
                              for the scene.  Must not be a null string.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   (None)
     Return Value:     The GB_id of the scene with the given label if it
                       exists
     Logic:
       1.  Call GB_get_id with the GB_SCENE type.
     End Header   ---------------------------------- */

GB_id GI_get_scene_id (label)
     char *label;
{
  return (GB_get_id (label, GB_SCENE));
}

/*   Function Header   -----------------------------
     Name:             GI_create_scene
     Description:      Create a new scene
     Parameters:       label: A character string containing the label of the
                              new scene.  Must not be a null string.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   A new scene is created if possible.
     Return Value:     The GB_id of the new scene.
     Logic:
       1.  Call GB_add_object to create the scene.
           * This uses all the defaults provided by the creator function
	     for scenes in DB.
     End Header   ---------------------------------- */

GB_id GI_create_scene (label)
     char *label;
{
  GS_scene_t new_scene;
  GB_id id;

  new_scene = (GS_scene_t) GB_add_object (label, GB_SCENE);
  id = GB_get_id (label, GB_SCENE);
  return (id);
}

/*   Function Header   -----------------------------
     Name:             GI_delete_scene
     Description:      Delete a scene.
     Parameters:       scene_id: scene ID.
     Preconditions:    The scene_id must be a valid id of a scene.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   The scene is deleted.
     Return Value:     (None)
     Logic:
       1.  Call GI_get_scene_internal.  This produces an error message
           if the scene does not exist.
       2.  Delete scene object.
     End Header   ---------------------------------- */

void GI_delete_scene (scene_id)
     GB_id scene_id;
{
  if (GI_get_scene_internal ("GI_delete_scene", scene_id))
    GB_delete_object (scene_id);
}



/*   Function Header   -----------------------------
     Name:             GI_update_scene
     Description:      Scene update function
     Parameters:       scene_id: scene ID
     Preconditions:    The scene_id must be a valid id of a scene.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   The scene has been updated.
     Return Value:     (None)
     Logic:
       1.  Retrieve the scene reporting errors.
       2.  If no errors, call GS_update_scene.
     End Header   ---------------------------------- */

void GI_update_scene (scene_id)
     GB_id scene_id;
{
  if (GI_get_scene_internal ("GI_update_scene", scene_id))
    GS_update_scene (scene_id);
}


/*   Function Header   -----------------------------
     Name:             GI_display_scene
     Description:      Display the scene on the given display
     Parameters:       display:  GB_id, Id of display to use.
                       scene_id: scene ID
		       x_offset: An interger to offset the scene from the
		                 left edge of the display.
		       y_offset: An interger to offset the scene from the
		                 top edge of the display.
     Preconditions:    The display must be a valid id of a display.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   The scene will be displayed on the given display.
     Return Value:     (None)
     Logic:
       1.  Retrieve the scene reporting errors.
       2.  If no errors, call GS_display_scene.
     End Header   ---------------------------------- */

void GI_display_scene (display, scene_id, x_offset, y_offset)
     GB_id display;
     GB_id scene_id;
     int x_offset, y_offset;
{
  if (display && GI_get_scene_internal ("GI_display_scene", scene_id))
    GS_display_scene (display, scene_id, x_offset, y_offset);
}


/*   Function Header   -----------------------------
     Name:             GI_get_scene_projection
     Description:      Fetch the projection of the given scene.
     Parameters:       scene_id: scene ID
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   (None)
     Return Value:     int, Projection of the scene or
                       0 if there is an error.
     Logic:
       1.  Retrieve the scene reporting errors.
       2.  Return the scene projection or 0.
     End Header   ---------------------------------- */

int GI_get_scene_projection (scene_id)
     GB_id scene_id;
{
  GS_scene_t scene;

  if (scene = GI_get_scene_internal ("GI_get_scene_projection",	scene_id))
    return (scene->projection);
  else
    return (0);
}


/*   Function Header   -----------------------------
     Name:             GI_set_scene_projection
     Description:      Set the scene's projection and parameters according
                       to the given projection descriptor.
     Parameters:       scene_id:   scene ID
                       projection: A projection descriptor.
     Preconditions:    Projection must be a projection descriptor.
                       The GeoData software must have been initialized prior
                       to the call.
		       The scene should exist.
     Postconditions:   The scene's projection and parameters are altered
                       according to the given projection descriptor.
     Return Value:     (None)
     Logic:
       1.  Retrieve the scene reporting errors.
       2.  Set the scene's projection and parameters.
     End Header   ---------------------------------- */

void GI_set_scene_projection (scene_id, projection)
     GB_id scene_id;
     GI_proj_desc *projection;
{
  GS_scene_t scene;
  int i;

  scene = GI_get_scene_internal ("GI_set_scene_projection", scene_id);
  if (scene && projection) {
    scene->projection = projection->code;
    for (i=0; i<NO_PROJECTION_PARAMS; i++)
      scene->parameters[i] = projection->param[i];
  }
}


/*   Function Header   -----------------------------
     Name:             GI_get_scene_projection_params
     Description:      Get the scene's current projection parameters.
     Parameters:       scene_id:   scene ID
                       parameters: pointer to an array of floats.  Must be
                                   at least NO_PROJECTION_PARAMS big.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
		       The Scene should exist.
     Postconditions:   The parameters array is filled with floats.
     Return Value:     (None)
     Logic:
       1.  Retrieve the scene reporting errors.
       2.  loop through all the parameters filling the return array.
     End Header   ---------------------------------- */

void GI_get_scene_projection_params (scene_id, parameters)
     GB_id scene_id;
     float parameters[];
{
  GS_scene_t scene;
  int i;

  scene = GI_get_scene_internal ("GI_get_scene_projection_params", scene_id);
  if (scene && parameters)
    for (i=0; i<NO_PROJECTION_PARAMS; i++)
      parameters[i] = scene->parameters[i];
}


/*   Function Header   -----------------------------
     Name:             GI_set_scene_projection_params
     Description:      Copies the given parameters into the scene.
     Parameters:       scene_id:   scene ID
                       parameters: pointer to an array of floats.  Must be
                                   at least NO_PROJECTION_PARAMS big.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
		       The Scene should exist.
     Postconditions:   The scenes parameter list is exacly the same as the
                       parameters given.
     Return Value:     (None)
     Logic:
       1.  Retrieve the scene reporting errors.
       2.  loop through all the parameters filling the scene.
     End Header   ---------------------------------- */

void GI_set_scene_projection_params (scene_id, parameters)
     GB_id scene_id;
     float parameters[];
{
  GS_scene_t scene;
  int i;

  scene = GI_get_scene_internal ("GI_set_scene_projection_params", scene_id);
  if (scene && parameters)
    for (i=0; i<NO_PROJECTION_PARAMS; i++)
      scene->parameters[i] = parameters[i];
}


/*   Function Header   -----------------------------
     Name:             GI_get_scene_projection_parameter
     Description:      Return a particular projection parameter.
     Parameters:       scene_id: scene ID
                       index:    integer index into the parameter array.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
		       The Scene should exist.
     Postconditions:   (None)
     Return Value:     float, the value of the indexed parameter
     Logic:
       1.  Check the index value.
       2.  Retrieve the scene reporting errors.
       3.  Return the indexed parameter.
     End Header   ---------------------------------- */

float GI_get_scene_projection_parameter (scene_id, index)
     GB_id scene_id;
     int index;        /* paremeter index: 0 - NO_PROJECTION_PARAMS */
{
  GS_scene_t scene;
  char message[100];
  float param=0.0;

  if ((index < 0) || (index >= NO_PROJECTION_PARAMS)) {
    sprintf(message, "Index out of range: %d", index);
    GB_message (GB_ERROR, "GI_get_scene_projection_parameter", message);
  }
  else {
    scene = GI_get_scene_internal ("GI_get_scene_projection_parameter",
				   scene_id);
    if (scene)
      param = scene->parameters[index];
  }
  return (param);
}


/*   Function Header   -----------------------------
     Name:             GI_set_scene_projection_parameter
     Description:      Set a particular projection parameter for a scene
     Parameters:       scene_id: scene ID
                       index:    integer index into the parameter array.
		       value:    float, new value for the indexed parameter.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
		       The Scene should exist.
     Postconditions:   The indexed parameter is set to the given value.
     Return Value:     (None)
     Logic:
       1.  Check the index value.
       2.  Retrieve the scene reporting errors.
       3.  Set the indexed parameter.
     End Header   ---------------------------------- */

void GI_set_scene_projection_parameter (scene_id, index, value)
     GB_id scene_id;
     int index;        /* paremeter index: 0 - NO_PROJECTION_PARAMS */
     float value;
{
  GS_scene_t scene;
  char message[100];

  if ((index < 0) || (index >= NO_PROJECTION_PARAMS)) {
    sprintf(message, "Index out of range: %d", index);
    GB_message (GB_ERROR, "GI_set_scene_projection_parameter", message);
  }
  else {
    scene = GI_get_scene_internal ("GI_set_scene_projection_parameter",
				   scene_id);
    if (scene)
      scene->parameters[index] = value;
  }
}


/*   Function Header   -----------------------------
     Name:             GI_add_scene_data
     Description:      Add a new data and rep pair to a scene
     Parameters:       scene_id: scene ID
                       rep_id:   representation ID
		       data_id:  data ID
		       priority: int, priority
     Preconditions:    Data, rep, and scene should exist.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   The data and rep are paired together in the scene's
                       rep list.
     Return Value:     (None)
     Logic:
       1.  Retrieve scene reporting errors.
       2.  Retrieve representation reporting errors.
       3.  Retrieve data set reporting errors.
       4.  Create and fill in a new rep-list entry.
       5.  Add it to the scene.
     End Header   ---------------------------------- */

void GI_add_scene_data (scene_id, rep_id, data_id, priority)
     GB_id scene_id, rep_id, data_id;
     int priority;
{
  GS_scene_t scene;
  GS_rep_list_t rep_list;
  GD_data_desc *data;
  GS_representation_t rep;

  scene = GI_get_scene_internal ("GI_add_scene_data", scene_id);
  rep = GI_get_rep_internal ("GI_add_scene_data", rep_id);
  data = (GD_data_desc *) GI_get_instance_internal ("GI_add_scene_data",
						    GB_DATA_DESC, data_id);
  if (scene && rep && data) {
    rep_list = (GS_rep_list_t) malloc (sizeof (GS_rep_list));
    rep_list->rep_id = rep_id;
    rep_list->data_id = data_id;
    rep_list->priority = priority;
    rep_list->next = scene->representations;
    scene->representations = rep_list;
  }
}


/*   Function Header   -----------------------------
     Name:             GI_delete_scene_data
     Description:      Remove a data and rep pair from the scene.
     Parameters:       scene_id: scene ID
                       rep_id:   representation ID
		       data_id:  data ID
     Preconditions:    Data, rep, and scene should exist.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   Data and rep pair are no longer together in the scene.
     Return Value:     (None)
     Logic:
       1.  Retrieve scene reporting errors.
       2.  Retrieve representation reporting errors.
       3.  Retrieve data set reporting errors.
       4.  Walk down the scene's rep-list looking for data and rep paired
           together.
       5.  Remove the rep-list entry if found.
     End Header   ---------------------------------- */

void GI_delete_scene_data (scene_id, rep_id, data_id)
     GB_id scene_id, rep_id, data_id;
{
  GS_scene_t scene;
  GS_rep_list_t rep_list, hat;
  GD_data_desc *data;
  GS_representation_t rep;

  scene = GI_get_scene_internal ("GI_delete_scene_data", scene_id);
  rep = GI_get_rep_internal ("GI_delete_scene_data", rep_id);
  data = (GD_data_desc *) GI_get_instance_internal ("GI_delete_scene_data",
						    GB_DATA_DESC, data_id);
  if (scene && rep && data) {
    hat = scene->representations;
    if ((hat->rep_id == rep_id) && (hat->data_id == data_id)) {
      scene->representations = hat->next;
      free (hat);
    }
    else {
      while (((hat->rep_id != rep_id) || (hat->data_id != data_id)) &&
	     (hat != NULL)) {
	rep_list = hat;
	hat = hat->next;
      }
      if (hat) {
	rep_list->next = hat->next;
	free (hat);
      }
    }
  }
}


/*   Function Header   -----------------------------
     Name:             GI_resize_scene
     Description:      Change the scene's size.
     Parameters:       scene_id: scene ID
                       width:    Int, new width, greater than 0.
                       height:   Int, new width, greater than 0.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
		       The scene should exist.
     Postconditions:   The scene's width and height have changed to the values
                       given.
     Return Value:     (None)
     Logic:
       1.  Check width and height parameters.
       2.  Retrieve the scene reporting errors.
       3.  Set width and height of scene.
     End Header   ---------------------------------- */

void GI_resize_scene (scene_id, width, height)
     GB_id scene_id;
     int width, height;
{
  GS_scene_t scene;
  char message[100];

  if ((1 > width) || (1 > height)) {
    sprintf(message, "New width %d, and height %d, not greater than 1.",
	    width, height);
    GB_message (GB_ERROR, "GI_resize_scene", message);
  }
  else {
    if (scene = GI_get_scene_internal ("GI_resize_scene", scene_id)) {
      scene->dimensions[0]=width;
      scene->dimensions[1]=height;
    }
  }
}

/*   Function Header   -----------------------------
     Name:             GI_get_scene_width
     Description:      Get the width of a scene
     Parameters:       scene_id: scene ID
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
		       The Scene should exist.
     Postconditions:   (None)
     Return Value:     int, width or
                       0 if an error.
     Logic:
       1.  Retrieve the scene reporting errors.
       2.  Return the width.
     End Header   ---------------------------------- */

int GI_get_scene_width (scene_id)
     GB_id scene_id;
{
  GS_scene_t scene;

  if (scene = GI_get_scene_internal ("GI_get_scene_width", scene_id))
    return (scene->dimensions[0]);
  else
    return (0);
}


/*   Function Header   -----------------------------
     Name:             GI_set_scene_width
     Description:      Set the width of a scene
     Parameters:       scene_id: scene ID
		       width:    int, new value for width.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
		       The Scene should exist.
     Postconditions:   The scene's width is the one given.
     Return Value:     (None)
     Logic:
       1.  Check the width value.
       2.  Retrieve the scene reporting errors.
       3.  Set the width value.
     End Header   ---------------------------------- */

void GI_set_scene_width (scene_id, width)
     GB_id scene_id;
     int width;
{
  GS_scene_t scene;
  char message[100];

  if (1 > width) {
    sprintf(message, "New width %d, is not greater than 1.", width);
    GB_message (GB_ERROR, "GI_set_scene_width", message);
  }
  else {
    scene = GI_get_scene_internal ("GI_set_scene_width", scene_id);
    if (scene)
      scene->dimensions[0] = width;
  }
}


/*   Function Header   -----------------------------
     Name:             GI_get_scene_height
     Description:      Get the height of a scene
     Parameters:       scene_id: scene ID
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
		       The Scene should exist.
     Postconditions:   (None)
     Return Value:     int, height or
                       0 if an error.
     Logic:
       1.  Retrieve the scene reporting errors.
       2.  Return the height.
     End Header   ---------------------------------- */

int GI_get_scene_height (scene_id)
     GB_id scene_id;
{
  GS_scene_t scene;

  if (scene = GI_get_scene_internal ("GI_get_scene_height", scene_id))
    return (scene->dimensions[1]);
  else
    return (0);
}


/*   Function Header   -----------------------------
     Name:             GI_set_scene_height
     Description:      Set the height of a scene
     Parameters:       scene_id: scene ID
		       height:   int, new value for height.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
		       The Scene should exist.
     Postconditions:   The scene's height is the one given.
     Return Value:     (None)
     Logic:
       1.  Check the height value.
       2.  Retrieve the scene reporting errors.
       3.  Set the height value.
     End Header   ---------------------------------- */

void GI_set_scene_height (scene_id, height)
     GB_id scene_id;
     int height;
{
  GS_scene_t scene;
  char message[100];

  if (1 > height) {
    sprintf(message, "New height %d, is not greater than 1.", height);
    GB_message (GB_ERROR, "GI_set_scene_height", message);
  }
  else {
    scene = GI_get_scene_internal ("GI_set_scene_height", scene_id);
    if (scene)
      scene->dimensions[1] = height;
  }
}


/*   Function Header   -----------------------------
     Name:             GI_get_scene_zoom
     Description:      Get the zoom factor for a scene
     Parameters:       scene_id: scene ID
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
		       The Scene should exist.
     Postconditions:   (None)
     Return Value:     int, zoom factor or
                       0 if an error.
     Logic:
       1.  Retrieve the scene reporting errors.
       2.  Return the zoom factor.
     End Header   ---------------------------------- */

int GI_get_scene_zoom (scene_id)
     GB_id scene_id;
{
  GS_scene_t scene;

  if (scene = GI_get_scene_internal ("GI_get_scene_zoom", scene_id))
    return (scene->zoom);
  else
    return (0);
}


/*   Function Header   -----------------------------
     Name:             GI_set_scene_zoom
     Description:      Set the zoom factor for a scene
     Parameters:       scene_id: scene ID
		       zoom:     int, new value for zoom.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
		       The Scene should exist.
     Postconditions:   The scene's zoom factor is the one given.
     Return Value:     (None)
     Logic:
       1.  Check the zoom value.
       2.  Retrieve the scene reporting errors.
       3.  Set the zoom value.
     End Header   ---------------------------------- */

void GI_set_scene_zoom (scene_id, zoom)
     GB_id scene_id;
     int zoom;
{
  GS_scene_t scene;
  char message[100];

  if ((1 > zoom) || (100 < zoom)) {
    sprintf(message, "New zoom %d, is not between 1 and 100.", zoom);
    GB_message (GB_ERROR, "GI_set_scene_zoom", message);
  }
  else {
    scene = GI_get_scene_internal ("GI_set_scene_zoom", scene_id);
    if (scene)
      scene->zoom = zoom;
  }
}


/*   Function Header   -----------------------------
     Name:             GI_get_scene_center
     Description:      Get the scene's current center.
     Parameters:       scene_id: scene ID
                       center:   pointer to an array of floats.  Must be
		                 at least three.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
		       The scene should exist.
     Postconditions:   The center array given is filled with floats.
     Return Value:     (None)
     Logic:
       1.  Retrieve the scene reporting errors.
       2.  loop through the center values filling the array.
     End Header   ---------------------------------- */

void GI_get_scene_center (scene_id, center)
     GB_id scene_id;
     float center[];
{
  GS_scene_t scene;
  int i;

  if (scene = GI_get_scene_internal ("GI_get_scene_center", scene_id)) {
    for (i=0; i<3; i++)
      center[i] = scene->center[i];
  }
}


/*   Function Header   -----------------------------
     Name:             GI_set_scene_center
     Description:      Set the scene's current center.
     Parameters:       scene_id: scene ID
                       center:   pointer to an array of floats.  Must be
		                 at least three.
     Preconditions:    The given array of floats should be full of values
                       from -1.0 to 0.0.
                       The GeoData software must have been initialized prior
                       to the call.
		       The scene should exist.
     Postconditions:   The center of the scene is set to the given array.
     Return Value:     (None)
     Logic:
       1.  Retrieve the scene reporting errors.
       2.  loop through the center values setting the scene's values.
     End Header   ---------------------------------- */

void GI_set_scene_center (scene_id, center)
     GB_id scene_id;
     float center[];
{
  GS_scene_t scene;
  char message[100];
  int i;

  if ((-1.0 > center[0]) || (1.0 < center[0]) ||
      (-1.0 > center[1]) || (1.0 < center[1]) ||
      (-1.0 > center[2]) || (1.0 < center[2])) {
    sprintf(message, "New Center %f, %f, %f, is not between -1.0 and 1.0.",
	    center[0], center[1], center[2]);
    GB_message (GB_ERROR, "GI_set_scene_center", message);
  }
  else {
    scene = GI_get_scene_internal ("GI_set_scene_center", scene_id);
    if (scene) {
      for (i=0; i<3; i++)
	scene->center[i] = center[i];
    }
  }
}


/*   Function Header   -----------------------------
     Name:             GI_get_scene_background
     Description:      Get the scene's current background.
     Parameters:       scene_id:   scene ID
                       background: pointer to an array of ints.  Must be
                                   at least three.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
		       The scene should exist.
     Postconditions:   The background array is set to the scene's current
                       background color.
     Return Value:     (None)
     Logic:
       1.  Retrieve the scene reporting errors.
       2.  loop through the background values filling the array.
     End Header   ---------------------------------- */

void GI_get_scene_background (scene_id, background)
     GB_id scene_id;
     int background[];
{
  GS_scene_t scene;
  int i;

  if (scene = GI_get_scene_internal ("GI_get_scene_background", scene_id)) {
    for (i=0; i<3; i++)
      background[i] = scene->background[i];
  }
}


/*   Function Header   -----------------------------
     Name:             GI_set_scene_background
     Description:      Set the scene's background color.
     Parameters:       scene_id:   scene ID
                       background: pointer to an array of ints.  Must be
		                   at least three.
     Preconditions:    The given array should be a valid color.  All values
                       from 0 to 255.
                       The GeoData software must have been initialized prior
                       to the call.
		       The scene should exist.
     Postconditions:   The background of the scene is set to the given color.
     Return Value:     (None)
     Logic:
       1.  Retrieve the scene reporting errors.
       2.  loop through the background values setting the scene's values.
     End Header   ---------------------------------- */

void GI_set_scene_background (scene_id, background)
     GB_id scene_id;
     int background[];
{
  GS_scene_t scene;
  char message[100];
  int i;

  if ((0 > background[0]) || (255 < background[0]) ||
      (0 > background[1]) || (255 < background[1]) ||
      (0 > background[2]) || (255 < background[2])) {
    sprintf(message, "New Background %d, %d, %d, is not between 0 and 255.",
	    background[0], background[1], background[2]);
    GB_message (GB_ERROR, "GI_set_scene_background", message);
  }
  else {
    scene = GI_get_scene_internal ("GI_set_scene_background", scene_id);
    if (scene) {
      for (i=0; i<3; i++)
	scene->background[i] = background[i];
    }
  }
}
