/*   File Header   ---------------------------------
     File Name:        cache.c
     Author and Date:  Paul B. Anderson, February 1991
     Description:      Reference data caching capabilities for the
                       Data Manager
     End Header   ---------------------------------- */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <cm/paris.h>
#include <cm/cm_file.h>
#include <geodata_private.h>

typedef struct GD_cache_field
{
  int                   last_use;
  int                   block_number;
  CM_field_id_t         field;
  struct GD_cache_field *next;
} GD_cache_field;

typedef struct GD_cache_entry 
{
  char file_name[GD_FILE_NAME_LENGTH];
  int  file_handle;
  int element_size;
  CM_vp_set_id_t vp_set;
  int  vp_ratio;
  struct GD_cache_field *fields;  
  struct GD_cache_entry *next;
} GD_cache_entry;

static int access_count;                    /*  Pseudo time for LRU  */
static GD_cache_entry *cache       = NULL;  /*  List of cache entries  */
static float cache_limit_percent   = 0.1;   /*  fraction of total memory
                                                usable for caching  */
static long int cache_size_in_bits = 0;     /*  cache usage in bits per
                                                processor  */


/*   Function Header   -----------------------------
     NAME:        GD_display_cache
     DESCRIPTION: take a data field and transform it into a transfer 
                  format field.
     PARAMETERS:  stream    a FILE* giving the output stream for
                            the cache structures
     PRECOND:     None
     POSTCOND:    No change in cache structures
     RETURN:      void
     LOGIC:
       1.  Output overall cache statistics.
       2.  Loop over all file entries in cache, displaying
           file level information.
       3.  For each file entry loop over all fields for the file,
           displaying the block number, field id and last use.
     End Header   ---------------------------------- */

void GD_display_cache(stream)
  FILE *stream;
{
  struct GD_cache_entry *entry_ptr;
  struct GD_cache_field *field_ptr;

  entry_ptr = cache;
  fprintf(stream,"Reference data cache limit percent: %f\n",cache_limit_percent);
  fprintf(stream,"          data cache size in bits:  %d\n",cache_size_in_bits);
  fprintf(stream,"          CM physical memory size:  %d\n",
          CM_physical_memory_limit);
  fprintf(stream,"Cache entries follow:\n");
  while (entry_ptr != NULL) {
    fprintf(stream,"  File %s, element size %d, vpset %d, vp ratio %d\n",
            entry_ptr->file_name,entry_ptr->element_size,
            entry_ptr->vp_set,entry_ptr->vp_ratio);
    field_ptr = entry_ptr->fields;
    fprintf(stream,"  Fields:\n");
    while (field_ptr != NULL) {
      fprintf(stream,"    Block number %d, field id %d, last use %d\n",
              field_ptr->block_number,field_ptr->field,field_ptr->last_use);
      field_ptr = field_ptr->next;
    }
    entry_ptr = entry_ptr->next;
  }
  fprintf(stream,"End of reference data cache\n");
}


/*   Function Header   -----------------------------
     NAME:        GD_purge_cache_to_limit
     DESCRIPTION: Purge cache holdings to the fraction of total processor
                  physical memory given by the variable cache_limit_percent
     PARAMETERS:  None
     PRECOND:     None
     POSTCOND:    Cache purged to specified limit (possibly except
                  for current return field which cannot be purged)
     RETURN:      void
     LOGIC:
       1.  Loop while the least recently accessed block is older than current
           and cache size exceeds the allowable limit.
       2.  Find the least recently accesed cache field.
       3.  If field is not current, purge it.  If removing the
           field leaves the file entry empty, close the file and purge
           the file entry also.
     End Header   ---------------------------------- */

void GD_purge_cache_to_limit()
{
  int access_time;
  struct GD_cache_entry *entry_ptr1, *entry_ptr2, *entry_ptr3, *entry_ptr4;
  struct GD_cache_field *field_ptr1, *field_ptr2, *field_ptr3, *field_ptr4;

  access_time = access_count-1;
  while((access_time < access_count) && 
        ((float)cache_size_in_bits/(float)CM_physical_memory_limit
                 >= cache_limit_percent)) {

    /* find least recently used cache holding  */
    access_time = access_count;
    entry_ptr1 = cache;  entry_ptr2 = NULL;  
    entry_ptr3 = NULL;   field_ptr3 = NULL;
    entry_ptr4 = NULL;   field_ptr4 = NULL;
    while (entry_ptr1 != NULL) {
      field_ptr1 = entry_ptr1->fields;  field_ptr2 = NULL;
      while (field_ptr1 != NULL) {
        if(field_ptr1->last_use < access_time) {
          access_time = field_ptr1->last_use;
          entry_ptr3 = entry_ptr1;  field_ptr3 = field_ptr1;
          entry_ptr4 = entry_ptr2;  field_ptr4 = field_ptr2;
        }
        field_ptr2 = field_ptr1;
        field_ptr1 = field_ptr1->next;
      }
      entry_ptr2 = entry_ptr1;
      entry_ptr1 = entry_ptr1->next;
    }

    if(access_time < access_count) {
      /*  purge cache holding  */
      CM_deallocate_heap_field(field_ptr3->field);
      cache_size_in_bits -= entry_ptr3->element_size*entry_ptr3->vp_ratio;
      if(field_ptr4 == NULL) {
        entry_ptr3->fields = field_ptr3->next;
        free(field_ptr3);
        if(entry_ptr3->fields == NULL) {
          /* remove cache entry */
          if(entry_ptr3->file_handle != -1) {
            CMFS_close(entry_ptr3->file_handle);
          }
          if(entry_ptr4 == NULL) {
            cache = entry_ptr3->next; }
          else {
            entry_ptr4->next = entry_ptr3->next;
          }
          free(entry_ptr3);
        } }
      else {
        field_ptr4->next = field_ptr3->next;
        free(field_ptr3);
      } /* if */
    } /* if */
  } /* while */
}


/*   Function Header   -----------------------------
     NAME:        GD_set_cache_limit
     DESCRIPTION: set the limit on available use of physical
                  memory for caching reference data
     PARAMETERS:  p         a float giving the fraction of physical
                            memory available for caching.  P must
                            lie in [0.0, 1.0]
     PRECOND:     None
     POSTCOND:    The cache has been purged to the specified limit
     RETURN:      void
     LOGIC:
       1.  If the supplied value is not within the prescribed
           range, return without changing cache status.
       2.  Otherwise, set cache_limit_percent and purge cache to
           the new limit value.
     End Header   ---------------------------------- */

void GD_set_cache_limit(p)
  float p;
{
  if(p >= 0.0 && p <= 1.0) {
/*    printf("\n\n**********   Cache before purge:\n");
    GD_display_cache(stdout);  */
    cache_limit_percent = p;
    access_count++;  /*  force all items to be eligible for purge  */
    GD_purge_cache_to_limit();
/*    printf("\n\n**********   Cache after purge:\n");
    GD_display_cache(stdout);  */
  }
}


/*   Function Header   -----------------------------
     NAME:        GD_get_cache_limit
     DESCRIPTION: Return the current fraction of physical
                  memory available for caching reference data
     PARAMETERS:  None
     PRECOND:     None
     POSTCOND:    No change
     RETURN:      double giving the cache fraction (between 0.0 and 1.0)
     LOGIC:
       1.  Return cache limit percent field.
     End Header   ---------------------------------- */

double GD_get_cache_limit()
{
  return cache_limit_percent;
}


/*   Function Header   -----------------------------
     NAME:        GD_get_ref_data
     DESCRIPTION: take a data field and transform it into a transfer 
                  format field.
     PARAMETERS:  file_name      A char * giving the data vault file name
                  block_number   An int giving the zero-relative block
                                 number (not the bit number) needed
                  size           An int giving the file data transfer
                                 size in bits
                  vp_set         VP set id for data in the file
                  gridded        An int set TRUE if the data set is
                                 gridded and FALSE otherwise
     PRECOND:     None
     POSTCOND:    Cache structures are updated, either by adjusting the
                  last use field of the returned field or by adding a
                  new field or file entry
     RETURN:      Field id containing the requested reference data
     LOGIC:
       1.  Loop over file entries for the requested file.  If not found,
           create a new file entry using supplied parameters.
       2.  Loop over all field entries for the requested file until
           the requested block is found.  If the field is not found,
           create a new entry for it, open the file if necessary, read
           in the data.
       3.  Update the last use field of the requested field, purge cache
           to its limit and return the field.
     End Header   ---------------------------------- */

CM_field_id_t GD_get_ref_data(file_name,block_number,size,vp_set,gridded)
  char *file_name; int block_number, size, gridded;  CM_vp_set_id_t vp_set;
{
  char message[128];
  struct GD_cache_entry *entry_ptr1, *entry_ptr2;
  struct GD_cache_field *field_ptr1, *field_ptr2;

  entry_ptr1 = cache;  entry_ptr2 = NULL;
  while (entry_ptr1 != NULL) {
    if(strncmp(file_name,entry_ptr1->file_name,GD_FILE_NAME_LENGTH-1) == 0) break;
    entry_ptr2 = entry_ptr1;
    entry_ptr1 = entry_ptr1->next;
  }
  if (entry_ptr1 == NULL) {
    entry_ptr1 = (struct GD_cache_entry *)malloc(sizeof(struct GD_cache_entry));
    if(entry_ptr2 == NULL) cache = entry_ptr1;
    else entry_ptr2->next = entry_ptr1;
    if(entry_ptr1 == NULL) {
      sprintf(message,"Unable to malloc space for cache entry");
/*      GB_message(GB_ERROR, "GD_get_ref_data", message);  */
      printf("%s\n",message);
      return(CM_no_field);
    }
    strncpy(entry_ptr1->file_name,file_name,GD_FILE_NAME_LENGTH-1);
    entry_ptr1->file_name[GD_FILE_NAME_LENGTH-1] = 0;
    entry_ptr1->file_handle = -1;
    entry_ptr1->element_size = size;
    entry_ptr1->vp_set = vp_set;
    entry_ptr1->vp_ratio = CM_geometry_total_vp_ratio(CM_vp_set_geometry(vp_set));
    entry_ptr1->fields = NULL;
    entry_ptr1->next = NULL;
  }

  field_ptr1 = entry_ptr1->fields;  field_ptr2 = NULL;
  while (field_ptr1 != NULL) {
    if(field_ptr1->block_number == block_number) break;
    field_ptr2 = field_ptr1;
    field_ptr1 = field_ptr1->next;
  }
  if (field_ptr1 == NULL) {
    field_ptr1 = (struct GD_cache_field *)malloc(sizeof(struct GD_cache_field));
    if(field_ptr2 == NULL) entry_ptr1->fields = field_ptr1;
    else field_ptr2->next = field_ptr1;
    if(field_ptr1 == NULL) {
      sprintf(message,"Unable to malloc space for cache field");
/*      GB_message(GB_ERROR, "GD_get_ref_data", message);  */
      printf("%s\n",message);
      return(CM_no_field);
    }
    field_ptr1->last_use = 0;
    field_ptr1->block_number = block_number;
    field_ptr1->field = CM_allocate_heap_field_vp_set(entry_ptr1->element_size,
                                                      entry_ptr1->vp_set);
    cache_size_in_bits += entry_ptr1->element_size*entry_ptr1->vp_ratio;

    /*  Read data vault file here, opening it if necessary  */
    if(entry_ptr1->file_handle == -1) {
      entry_ptr1->file_handle = CMFS_open(entry_ptr1->file_name,CMFS_O_RDONLY, 0644);
      if (entry_ptr1->file_handle == -1){ 
	CM_perror("Opening reference data file"); 
	printf("%s",entry_ptr1->file_name); 
	exit(1); }
    }
    if(CMFS_lseek(entry_ptr1->file_handle,entry_ptr1->element_size*block_number,
                  CMFS_L_SET) == -1)
      { CMFS_perror("error seeking for reference data"); return(0); }
    /* read a block */
    if(CMFS_read_file_always(entry_ptr1->file_handle,field_ptr1->field,
                             entry_ptr1->element_size) != entry_ptr1->element_size)
      { CMFS_perror("error reading reference data"); return(CM_no_field); }
	
    if(!gridded) {
      CMFS_close(entry_ptr1->file_handle);
      entry_ptr1->file_handle = -1;
    }
    field_ptr1->next = NULL;
  }

  field_ptr1->last_use = ++access_count;
  GD_purge_cache_to_limit();

  return field_ptr1->field;

}

