/*   File Header   ---------------------------------

     File Name:       expand.c

     Author and Date: George Lamb--7/10/90

     Description: functions to expand loaded reference data into
     geodata internal format.  The load functions know the stored format,
     the internal format is:
       points & meshes = (lat, lon, alt, val)
       vectors = (lat, lon, alt, val, lat, lon, alt, val)
     where all units are specified in radians and meters.
     All functions return 1 for sucess, 0 for failure.
      
     functions defined:
        int GD_expand_2dp();
	int GD_expand_3dp();
	int GD_expand_2dp_nv();
	int GD_expand_3dp_nv();
	int GD_expand_2dv();
	int GD_expand_3dv();
	int GD_expand_2dv_nv();
	int GD_expand_3dv_nv();
	int GD_expand_wdb2();
	int GD_expand_3d_vchain();
	int GD_expand_3d_vchain_nv();
	int GD_expand_mesh_32();
	int GD_expand_dbdb5();

     End Header   ---------------------------------- */

#include <geodata.h>
#include <geodata_private.h>
#include <stdio.h>
#include <math.h>
#include <cm/paris.h>

char note[64];

int GD_expand_2dp(dest, src, desc)
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
/*  printf("GD_expanding %d, %d, %d, %d.\n", dest, src, desc); */
}


/* 3dp format:  (x, y, z, v) or (lat, lon, alt, v) */
int GD_expand_3dp(dest, src, desc)       
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
  if (desc->coord_syst == GD_CART_3D){
    GD_map_3d_angular(dest, src, desc->origin);
    CM_u_move_1L(dest+3*32, src+3*32, 32);
  }
  else {
    CM_u_move_1L(dest+3*32, src+3*32, 32);
  }
}     /* end of GD_expand_3dp */



int GD_expand_2dp_nv(dest, src, desc)
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
  GB_message(GB_ERROR, "GD_expand_2dp_nv", "undefined expansion");
  return(0);
}

int GD_expand_3dp_nv(dest, src, desc)
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
  GB_message(GB_ERROR, "GD_expand_3dp_nv", "undefined expansion");
  return(0);
}


/* 2dv format:  (x, y, x, y, v) or (lat, lon, lat, lon, v).
   examples are shorelines... */
int GD_expand_2dv(dest, src, desc)       
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
  GD_point xp, yp;
  float d;
  
  if (desc->coord_syst == GD_CART_2D) {
    d = compute_axes(desc->r, desc->nv, desc->yp, &xp, &yp);
    printf("xp = (%f %f %f), yp = (%f %f %f).\n", xp.lat, xp.lon, 
	   xp.alt, yp.lat, yp.lon, yp.alt);
    GD_map_2d_plane_3d(dest, dest, dest+32, &xp, &yp, desc->r);
    GD_map_2d_plane_3d(dest+4*32, dest+4*32, dest+5*32, &xp, &yp, desc->r);
    CM_u_move_1L(dest+3*32, src+4*32, 32);          /* val */
    GD_map_3d_angular(dest, dest, desc->origin);
    GD_map_3d_angular(dest+4*32, dest+4*32, desc->origin);
    CM_u_move_1L(dest+7*32, dest+3*32, 32);         /* val */
  }
  else if (desc->coord_syst == GD_MAP){
    switch (desc->proj_fn) {
    case GS_NO_PROJECTION:
      CM_f_move_constant_1L(dest+2*32, 0.0, 23, 8);   /* alt */
      CM_u_move_1L(dest+3*32, src+4*32, 32);          /* val */
      CM_f_move_constant_1L(dest+6*32, 0.0, 23, 8);   /* alt */
      CM_u_move_1L(dest+7*32, dest+3*32, 32);         /* val */
      break;
    default: 
      sprintf(note, "inverse projection function %d not implemented.", 
	      desc->proj_fn);
      GB_message(GB_NOTIFY, "GD_expand_2dv", note);
      break;
    }
  }
  else {
    GB_message(GB_ERROR, "GD_expand_2dv", "2dv coord_syst mismatch");
    return(0);
  }
  return(1);
}   /* end of GD_expand_2dv */



int GD_expand_3dv(dest, src, desc)       
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
  GB_message(GB_ERROR, "GD_expand_3dv_nv", "undefined expansion");
  return(0);
}

int GD_expand_2dv_nv(dest, src, desc)
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
  GB_message(GB_ERROR, "GD_expand_2dv_nv", "undefined expansion");
  return(0);
}

int GD_expand_3dv_nv(dest, src, desc)
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
  GB_message(GB_ERROR, "GD_expand_3dv_nv", "undefined expansion");
  return(0);
}

int GD_expand_wdb2(dest, src, desc)      
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
  CM_set_context();

  /*  Fill in Altitude and value fields with zeros  */
  CM_f_move_zero_always_1L(CM_add_offset_to_field_id(dest,64),23,8); 
  CM_f_move_zero_always_1L(CM_add_offset_to_field_id(dest,96),23,8);
  CM_f_move_zero_always_1L(CM_add_offset_to_field_id(dest,192),23,8);
  CM_f_move_zero_always_1L(CM_add_offset_to_field_id(dest,224),23,8);

  /*  Fill in second coordinate of vector  */
  CM_f_move_always_1L(CM_add_offset_to_field_id(dest,128),
                      CM_add_offset_to_field_id(src,0),23,8);
  CM_f_move_always_1L(CM_add_offset_to_field_id(dest,160),
                      CM_add_offset_to_field_id(src,32),23,8);

  /*  Get coordinates from previous element  */
  CM_get_from_news_always_1L(CM_add_offset_to_field_id(dest,0),
                             CM_add_offset_to_field_id(src,0),
                             0,CM_downward,64);

  /*  Select only valid processors whose command is 'DRAW'  */
  CM_u_eq_constant_1L(CM_add_offset_to_field_id(src,64),2,2);
  CM_logand_context_with_test();

  /* evil hack*/
/*  {
    CM_field_id_t temp=CM_allocate_stack_field(1);
    CM_logxor_3_1L (temp,dest+63,dest+128+63,1);
    CM_logxor_2_1L (CM_context_flag,temp,1);
    CM_deallocate_stack_through (temp);
  }
*/
/*  printf("%d WDB2 vectors selected; Statistics before projection follow:\n",
         CM_global_count_context());*/
  return(0);
}


int GD_expand_3d_vchain(dest, src, desc) 
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
  GB_message(GB_ERROR, "GD_expand_3d_vchain", "undefined expansion");
  return(0);
}

int GD_expand_3d_vchain_nv(dest, src, desc)
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
  GB_message(GB_ERROR, "GD_expand_3d_vchain_nv", "undefined expansion");
  return(0);
}

/* the context here will be selectively axed */

int GD_expand_mesh_32(dest, src, desc)      
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
  float space[3];
  int j;
  CM_geometry_id_t geom = CM_vp_set_geometry(CM_current_vp_set);

  for (j=0; j < desc->rank; j++)
    space[j] = desc->extent[j] / CM_geometry_axis_length(geom, j);

  /* dest = index * spacing[i]; */
  CM_set_context();


  CM_my_news_coordinate_1L (dest+96,0,32);
  CM_f_u_float_2_2L (dest,dest+96,32,23,8);
  CM_f_mult_const_add_const_1L (dest, dest, space[0], desc->origin[0], 23, 8);



  CM_my_news_coordinate_1L (dest+96,1,32);
  CM_f_u_float_2_2L (dest+32,dest+96,32,23,8);
  CM_f_mult_const_add_const_1L (dest+32, dest+32, space[1], desc->origin[1],
				23, 8);


  if (desc->rank == 3){
    CM_my_news_coordinate_1L (dest+96,2,32);
    CM_f_u_float_2_2L (dest+64,dest+96,32,23,8);
    CM_f_mult_const_add_const_1L (dest+64, dest+64, space[2], desc->origin[2],
				  23, 8);
  }
  else CM_f_move_zero_1L (dest+64, 23, 8);      

  if(src != CM_no_field) CM_u_move_1L(dest+96, src, 32);
  if (desc->print_stats){
    printf ("expand mesh 32\n");
    GD_transfer_statistics (4,dest);
  }
}   /* end of GD_expand_mesh_32 */


int GD_expand_dbdb5(dest, src, desc)
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
  float space[3];
  int j;
  CM_geometry_id_t geom = CM_vp_set_geometry(CM_current_vp_set);
  CM_field_id_t temp, ctx;

  for (j=0; j < desc->rank; j++)
    space[j] = desc->extent[j] /
      ( CM_geometry_axis_length(geom, j) * desc->span[j]);
  
  ctx = CM_allocate_stack_field(1);
  CM_store_context(ctx);
  /* dest = index * spacing[i] + origin[i] */
  CM_my_news_coordinate_1L (dest+32,0,32);
  CM_f_u_float_2_2L (dest,dest+32,32,23,8);
  CM_f_mult_const_add_const_1L (dest, dest, space[0], desc->origin[0], 23, 8);

  CM_my_news_coordinate_1L (dest+64,1,32);
  CM_f_u_float_2_2L (dest+32,dest+64,32,23,8);
  CM_f_mult_const_add_const_1L (dest+32, dest+32, space[1], desc->origin[1], 23, 8);

  /* for dbdb5 data, alt = value */
  CM_u_move_zero_1L(dest+64, 32);
  if(src != CM_no_field) {
    temp = CM_allocate_stack_field(16); /***dest & src cannot overlap***/
    CM_u_move_1L(temp, src, 16); 
    CM_f_s_float_2_2L (dest+64, temp, 16, 23, 8);
    CM_f_add_constant_2_1L(dest+64, desc->origin[2], 23 ,8); 
  }
  CM_f_move_1L(dest+96, dest+64, 23, 8);
  CM_load_context(ctx);

  CM_f_multiply_constant_2_1L (dest+64,desc->exaggeration,23,8);
  CM_deallocate_stack_through(ctx);
}   /* end of expand_dbdb5 */

#define MAGIC 22.5

int GD_expand_dbdb5_edges (dest, src, desc)
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
  CM_field_id_t x = CM_allocate_stack_field(32);
  CM_field_id_t y = CM_allocate_stack_field(32);
  CM_field_id_t temp = CM_allocate_stack_field(32);

  CM_set_context();

  CM_f_s_float_2_2L (dest+64,src,16,23,8);
  CM_f_s_float_2_2L (dest+96,src,16,23,8);

  CM_my_news_coordinate_1L (x,0,32);
  CM_my_news_coordinate_1L (y,1,32);

  CM_load_context(y+8);
  CM_invert_context();
  CM_f_move_constant_1L (dest,0.0,23,8);
  CM_logand_context(y);
  CM_f_move_constant_1L (dest,MAGIC/256.0,23,8);
  CM_u_move_1L (dest+31,y+3,1);
  CM_load_context(y+8);
  CM_invert_context();
  
  CM_f_u_float_2_2L (temp,y+1,3,23,8);
  CM_f_mult_const_add_1L (dest,temp,MAGIC,dest,23,8);
  
  CM_f_u_float_2_2L (temp,y+4,4,23,8);
  CM_f_multiply_constant_2_1L (temp,MAGIC,23,8);
  CM_f_u_float_2_2L (dest+32,x,32,23,8);
  CM_f_mult_const_add_1L (dest+32,dest+32,MAGIC/256.0,temp,23,8);


  CM_load_context (y+8);  
  CM_u_subfrom_constant_2_1L (x,255,32);
  CM_f_move_constant_1L (dest+32,0.0,23,8);
  CM_load_context(y);
  CM_logand_context (y+8);
  CM_f_move_constant_1L (dest+32,MAGIC/256.0,23,8);
  CM_u_move_1L (dest+32+31,y+4,1);
  CM_load_context (y+8);  
  CM_f_u_float_2_2L (temp,y+1,4,23,8);
  CM_f_mult_const_add_1L (dest+32,temp,MAGIC,dest+32,23,8);
  
  CM_f_u_float_2_2L (temp,y+5,3,23,8);
  CM_f_multiply_constant_2_1L (temp,MAGIC,23,8);
  CM_f_u_float_2_2L (dest,x,32,23,8);
  CM_f_mult_const_add_1L (dest,dest,MAGIC/256.0,temp,23,8);

  CM_set_context();
  CM_f_subfrom_constant_2_1L (dest,90.0,23,8);
  CM_f_multiply_constant_2_1L (dest+64,desc->exaggeration,23,8);
  
  if (desc->context == 0) desc->context = CM_allocate_heap_field(1);
  CM_lognot_2_1L (desc->context,y,1);
  CM_deallocate_stack_through(x);
} 

int GD_expand_border (dest,src,desc)
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
}


int GD_expand_client(dest, src, desc)      
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
  int coordinate_rank;

  if(desc->type == GD_MESH) {
    GD_expand_mesh_32(dest,src,desc); }
  else {
    coordinate_rank = 3;
    if(desc->coord_syst == GD_CART_2D || 
       desc->coord_syst == GD_MAP) coordinate_rank = 2;
    CM_u_move_1L(dest,src,32*coordinate_rank);
    if(coordinate_rank == 2) CM_f_move_zero_1L(dest+64,32);
    if(desc->type == GD_VECTOR) {
      CM_u_move_1L(dest+128, src+32*coordinate_rank, 32*coordinate_rank);
      if(coordinate_rank == 2) CM_f_move_zero_1L(dest+128+64,32);
      CM_u_move_1L(dest+96,  src+2*32*coordinate_rank, 32);
      CM_u_move_1L(dest+224, src+2*32*coordinate_rank, 32);
    }
    else {
      CM_u_move_1L(dest+96,  src+coordinate_rank*32, 32);
    }
  }
  CM_f_multiply_constant_2_1L (dest+64,desc->exaggeration,23,8);
}   /* end of GD_expand_client */


int GD_expand_dlg(dest, src, desc)      
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
  CM_set_context();
  
  CM_f_move_zero_always_1L(CM_add_offset_to_field_id(dest,128+96),23,8);
  CM_f_move_zero_always_1L(CM_add_offset_to_field_id(dest,96),23,8);
  
  CM_u_move_always_1L(CM_add_offset_to_field_id(dest,128),src,96);
  CM_get_from_news_always_1L(dest,src,0,CM_downward,96);
  
  CM_swap_2_1L (dest+64,dest+96,32);
  CM_swap_2_1L (dest+64+128,dest+96+128,32);
  CM_logand_context(CM_add_offset_to_field_id(src,96));
  return(0);
}



static GD_load_functions load_functions[]={
  {3*32+1, GD_expand_2dp},
  {4*32+1, GD_expand_3dp},
  {2*32+1, GD_expand_2dp_nv},
  {3*32+1, GD_expand_3dp_nv},
  {5*32+1, GD_expand_2dv},
  {7*32+1, GD_expand_3dv},
  {4*32+1, GD_expand_2dv_nv},
  {5*32+1, GD_expand_3dv_nv},
  {2*32+2, GD_expand_wdb2},
  {4*32+2, GD_expand_3d_vchain},
  {3*32+2, GD_expand_3d_vchain_nv},
  {32, GD_expand_mesh_32},
  {16, GD_expand_dbdb5},
  {16, GD_expand_dbdb5_edges},
  {4*32, GD_expand_client},
  {7*32, GD_expand_client},
  {97, GD_expand_dlg},
  {1, GD_expand_border},
};


int GD_check_bad_type(type, funct)
     int type;
     char funct[];
{
  char note[64];
  if (!((type >GD_BEGINNING_OF_LOAD_TYPES) && (type < GD_END_OF_LOAD_TYPES))){
    sprintf(note, "Type %d is out of range.", type);
    GB_message(GB_ERROR, funct, note);   return(1);
  }
  else return(0);
}


int GD_sizeof_type(type)
     int type;
{
  if (GD_check_bad_type(type, "GD_sizeof_type")) return(0);
  else return(load_functions[type-GD_2DP].size);
}    /* end of GD_sizeof_type */



int GD_expand_type(type, dest, src, desc)
     int type;
     CM_field_id_t dest, src;
     GD_data_desc *desc;
{
  if (GD_check_bad_type(type, "GD_expand_type")) return(0);
  else return((*(load_functions[type-GD_2DP].expand_func)) 
	 (dest, src, desc));
}
