/*   File Header   ---------------------------------

     File Name:       interface.c

     Author and Date: George Lamb--3/19/90
                      Paul B. Anderson  12/90 (adapted from source.c)

     Description:     this file contains code to implement the 
     scene generator and client interface to the data manager.
     The functions implemented are:

     void          GD_open_ds(dsid, projection,parameters,block_dimensions)
     CM_field_id_t GD_read_ds(dsid, block)
     void          GD_close_ds(dsid)
    End Header   ---------------------------------- */


#include <geodata.h>
#include <geodata_private.h>
#include <cm/paris.h>
#include <cm/cm_file.h>
#include <stdio.h>
#include <math.h>

extern void GD_open_ref_source();
extern void GD_make_transfer();
extern void GD_mv_fe_to_cm();
extern CM_field_id_t GD_read_ref_source();
extern CM_field_id_t GD_cmalloc_std ();

/* change fortran CM geometry to C CM geometry */
void GD_fort_to_c_geo(desc)
     GD_data_desc *desc;
{
  struct CM_axis_descriptor axis_0,axis_1;
  CM_axis_descriptor_t axes[2];
  
  axes[0]=(&axis_0);
  axes[1]=(&axis_1);
  bcopy ((desc->fortran_geometry->descriptors)[1],&axis_0,
	 sizeof(struct CM_axis_descriptor));
  if(desc->rank==2)
    bcopy ((desc->fortran_geometry->descriptors)[2],&axis_1,
	   sizeof(struct CM_axis_descriptor));
  axis_0.weight=0;
  axis_1.weight=0;
  CM_set_vp_set_geometry(*(desc->vp_set),
			 CM_create_detailed_geometry(axes,desc->rank));
}

void GD_reset_to_fort_geo(desc)
     GD_data_desc *desc;
{
  CM_set_vp_set_geometry(*(desc->vp_set),desc->fortran_geometry);
}

/*   Function Header   -----------------------------
     NAME:        GD_next_block
     DESCRIPTION: Count up the block array using the mixed
                  radix numbering system determined by the
                  shape of a gridded file
     PARAMETERS:  block  An integer array of block numbers (zero
                         relative)
                  span   An integer array giving the shape of the
                         gridded data set
     PRECOND:     Block should be a legal value
     POSTCOND:    Block is modified to the successor value in the
                  mixed radix system
     RETURN:      None
     LOGIC: 1.  Iterate from the least significant digit to the
                most:
            2.  Increment digit.
            3.  If digit is equal to the corresponding span value,
                set it to 0 and continue in next position.  Otherwise
                update is complete.
     End Header ---------------------------------- */

int GD_next_block(block,span)
  int block[], span[];
{
  int i, at_end;

  for(at_end = TRUE, i=2; i>=0; i--) {
    if(++block[i] == span[i]) block[i] = 0;
    else { at_end = FALSE;  break; }
  }

  return at_end;
}

/*   Function Header   -----------------------------
     NAME:        GD_block_number
     DESCRIPTION: print transfer statistics for the current
                  data set
     PARAMETERS:  block  An integer array of block numbers (zero
                         relative)
                  span   An integer array giving the shape of the
                         gridded data set
     PRECOND:     Block array has a legal value
     POSTCOND:    None
     RETURN:      Integer block number (zero relative)
     LOGIC: 1.  Iterate over the digits of block from most signigifcant
                to least:
            2.  Evaluate indexing polynomial.
            3.  Return final block number value.
     End Header ---------------------------------- */

int GD_block_number(block,span)
  int block[], span[];
{
  int i, block_no;

  for(block_no = block[0], i=1; i<3; i++) {
    block_no = block_no*span[i] + block[i];
  }

  return block_no;
}

/*   Function Header   -----------------------------
     NAME:        GD_compute_origin
     DESCRIPTION: Compute coordinate system origin for a given
                  block of a gridded data set
     PARAMETERS:  origin A float array returned with the coordinate
                         system origin of the gridded data set cell 
                         indicated by the block array
                  base   A float array giving the origin of the
                         first cell of a gridded data set
                  incr   A float array giving the origin increments of
                         the gridded data set
                  wrap   A float array giving the coordinate system
                         wrap values (0.0 if no wrap) for the gridded 
                         data set
                  block  An integer array of block numbers (zero
                         relative)
     PRECOND:     Base, incr, wrap, and block are defined
     POSTCOND:    Origin is updated
     RETURN:      None
     LOGIC: 1.  Iterate over the fields of block from most significant
                to least:
            2.  Compute the new origin ignoring wrap considerations.
            3.  If the wrap value is positive and the origin exceeds
                the wrap value, map the origin back into the proper
                range.
     End Header ---------------------------------- */

void GD_compute_origin(origin,base,incr,wrap,block)
  float origin[], base[], incr[], wrap[];
  int block[];
{
  int i, block_no;

  for(i=0; i<3; i++) {
    origin[i] = base[i] + block[i] * incr[i];
    if(wrap[i] > 0.0 && origin[i] >= wrap[i]) origin[i] -= wrap[i]+wrap[i];
  }
}

/*   Function Header   -----------------------------
     NAME:        GD_transfer_statistics
     DESCRIPTION: print transfer statistics for the current
                  data set
     PARAMETERS:  k     The number of fields to dump
                  src  The field id containing the transfer
     PRECOND:     none
     POSTCOND:    none
     RETURN:      none
     LOGIC: 1.  Iterate over the k fields
            2.  Compute the min, max, mean, and variance of
                each field and output as a table.
     End Header ---------------------------------- */

void GD_transfer_statistics(k, src)
  int k; CM_field_id_t src;
{
  long i, n;  CM_field_id_t temp1, f;
  double min, max, mean, var;

  temp1 = CM_allocate_stack_field(32);
  n = CM_global_count_context();

  printf("Transfer Field Statistics: %d active elements\n",n);
  if(n > 0) {
    printf("        Minimum      Maximum         Mean     Variance\n");

    for (i = 0; i < k; i++) {
      f = CM_add_offset_to_field_id(src, 32*i);
      min = CM_global_f_min_1L(f,23,8);
      max = CM_global_f_max_1L(f,23,8);
      if(n > 1) {
        mean = CM_global_f_add_1L(f,23,8)/ (double) n;
        CM_f_multiply_3_1L(temp1,f,f,23,8);
	var = 0.0;
/*        var = ((double) CM_global_f_add_1L(temp1,23,8) - n * mean * mean) / 
              (double) (n - 1); 
*/
      }
      else { mean = min;  var = 0.0; }
      printf("%2d   %10g   %10g   %10g   %10g\n",i,min,max,mean,var);
    } 
  }

  CM_deallocate_stack_through(temp1);

}  /*  GD_transfer_statistics  */



/*   Function Header   -----------------------------
     NAME:        GD_trim_mesh_edges
     DESCRIPTION: remove final elements of non-periodic axes
     PARAMETERS:  ds              data set descriptor
     PRECOND:     none
     POSTCOND:    the current context reflects the change.
     RETURN:      none
     LOGIC: 1.  iterate over the axes
            2.  if non-periodic, turn off the last element
     End Header ---------------------------------- */

void GD_trim_mesh_edges (ds)
     GD_data_desc *ds;
{
  CM_geometry_id_t geometry = CM_vp_set_geometry(CM_current_vp_set);
  CM_field_id_t temp = CM_allocate_stack_field(32);
  int i;

  for (i=0;i<CM_geometry_rank (geometry);i++)
    if (ds->periodic_axes[i] == GD_NON_PERIODIC){
      CM_my_news_coordinate_1L (temp,i,32);
      CM_u_lt_constant_1L (temp,CM_geometry_axis_length(geometry,i)-1,32);
      CM_logand_context_with_test();
    }
  CM_deallocate_stack_through (temp);
  if (ds->print_stats){
    printf ("trim mesh edges\n");
    GD_transfer_statistics((ds->type == GD_VECTOR)?8:4,ds->transfer);
  }
}


/*   Function Header   -----------------------------
     NAME:        GD_open_ds
     DESCRIPTION: open a data set for retrieving data
     PARAMETERS:  dsid              data id being retrieved
                  projection        projection code to use
                  parameters        projection parameters
                  block_dimensions  grid dimensions
     PRECOND:     none
     POSTCOND: a new source is allocated which contains information
     needed to read the data, and the max dimensions needed to fill the
     geowin is returned to the caller.
     RETURN:      a ptr to a data source
     LOGIC: 1.  retrieve the data descriptor for the id, if NULL, return NULL
            2.  allocate a new data source
	    3.  compute the dims needed to fill the geowin
	    4.  if desc == NULL, create the return field in transfer
	    format, otherwise try and match the descriptor.  Units, coordsyst,
	    type, fe_data, extent & shape appropriately interpolated.
            5.  return a ptr to the source for future access, and the
	    dims for user reference.
     End Header ---------------------------------- */

int *GD_open_ds (desc,projection,parameters,block_dimensions)
     GD_data_desc *desc;
     int   projection;
     float parameters[];
     int   block_dimensions[];
{
  int j;

  block_dimensions[0] = 0;
  block_dimensions[1] = 0;
  block_dimensions[2] = 0;

  if(desc->fortran_data_flag)GD_fort_to_c_geo(desc);

  desc->at_end = FALSE;  
  desc->file = -1;
  if(*(desc->vp_set) == -1)
    *(desc->vp_set) = CM_allocate_vp_set
      (CM_create_geometry(desc->geometry,desc->rank));
  CM_set_vp_set(*(desc->vp_set));
  CM_set_context(); 
  desc->save_origin[0]=desc->origin[0];
  desc->save_origin[1]=desc->origin[1];
  desc->save_origin[2]=desc->origin[2];
}   /* GD_open_ds */


/*   Function Header   -----------------------------
     NAME:        GD_read_ds
     DESCRIPTION: read data for the given data set, put it in a standard
                  transfer format, and return a cm ptr to it.
     PARAMETERS:  dsid        data set id to read
                  block       3 element array of ints returned with the
                              current block coordinates
                  projection  projection code to use
                  parameters  projection parameters
     PRECOND:     the data set has been opened
     POSTCOND:    a new field is allocated on the heap containing data
                  from the given data set, in 3d standard coordinates.
     RETURN: the field id for the next block of the data set, translated to
             std format and clipped appropriately, 0 => no more blocks
     LOGIC: 1. if we're using block data, see if we're at the last block
            2. check if the user has his own desired descriptor, and match it.
	    3. check to see if we're using refrence data, if so pass it on
	    4. else...allocate a std transfer field, and fill it in.
     End Header ---------------------------------- */

CM_field_id_t GD_read_ds (desc,projection,parameters,block)
     GD_data_desc *desc;
     int   projection;
     float parameters[];
     int block[];
{
  int n,i,j,bits, any_visible;  char note[32];
  float wrap[3];  CM_field_id_t success;
  float block_incr[3];
  float temp_origin;

  if(desc->at_end) return(CM_no_field);

  CM_set_vp_set(*(desc->vp_set));
  CM_set_context(); 
  /* handle gridded reference data */

  if(!((desc->span[0] == 1)&&(desc->span[1] == 1) && (desc->span[2] == 1))){
    block_incr[0]=desc->extent[0]/desc->span[0];
    block_incr[1]=desc->extent[1]/desc->span[1];
    block_incr[2]=desc->extent[2]/desc->span[1];
    bits = GD_sizeof_type(desc->load_type);
    /*  For grids, materialize the coordinate mesh and check visibility */
    if(desc->type == GD_MESH) {
      for(i=0; i<3; i++) {
	switch (desc->units[i]) {
	case GD_RADIANS:  wrap[i] = M_PI;
	  break;
	case GD_DEGREES:  wrap[i] = 180.0;
	  break;
	default:          wrap[i] = 0.0;
	}
      }
      desc->transfer = GD_cmalloc_std(desc->type,1);
      success = CM_add_offset_to_field_id(desc->transfer,96);
      any_visible = FALSE;
      while((!any_visible) && (!desc->at_end)) {
	GD_compute_origin(desc->origin,desc->save_origin,
			  block_incr,wrap,block);
	GD_expand_mesh_32(desc->transfer,CM_no_field,desc);
	GD_convert_units_std(desc->transfer,desc->transfer,desc,0);
	GD_project(projection,desc->transfer,success,parameters,
		   desc->coord_syst,desc->origin);
	any_visible = CM_global_logior_1L(success,1);
	if(!any_visible) desc->at_end = GD_next_block(block,desc->span);
      }
    }
    if(desc->at_end) return(CM_no_field);   /*  This isn't nice  */
    /*  get reference data from file based on block array */
    n = GD_block_number(block,desc->span);
    desc->values = GD_get_ref_data(desc->filename,n,bits,
				   *(desc->vp_set),TRUE);
    /*  build and return the standard transfer field  */
    GD_make_transfer(desc->transfer, desc, projection, parameters);
    desc->at_end = GD_next_block(block,desc->span);
  }
  /*  Handle ungridded reference data  */
  else {
    bits = GD_sizeof_type(desc->load_type);
    /* bless me father for I have sinned...*/
    if(desc->cm_fe_flag) GD_mv_fe_to_cm(desc);
    if (!desc->values)
      desc->values = GD_get_ref_data(desc->filename,0,bits,
				     *(desc->vp_set),FALSE);
    desc->transfer = GD_cmalloc_std(desc->type,1);
    GD_make_transfer(desc->transfer, desc, projection, parameters);
    for(i=0; i<3; i++) block[i] = 1;
    desc->at_end = TRUE;  /* For next time */ 
  }
  if (desc->type == GD_MESH) GD_trim_mesh_edges (desc);
  if(desc->context != NULL) CM_logand_context (desc->context);
  if (desc->print_stats){
    printf ("transfer field\n");
    GD_transfer_statistics((desc->type == GD_VECTOR)?8:4,desc->transfer);
  }
  return (desc->transfer);
}   /* GD_read_ds */


/*   Function Header   -----------------------------
     NAME:        GD_close_ds
     DESCRIPTION: close a data set 
     PARAMETERS:  dsid -- the data set in question
     PRECOND:     the data set has been previously opened
     POSTCOND:    transfer field for a client data set has been
                  released
     RETURN:      none
     LOGIC: 1.  If field is a client field, release transfer field
            2.  Otherwise TBD.
     End Header ---------------------------------- */

void GD_close_ds(desc)
     GD_data_desc *desc;
{
  if(desc->fortran_data_flag)GD_reset_to_fort_geo(desc);

  desc->origin[0]= desc->save_origin[0];
  desc->origin[1]= desc->save_origin[1];
  desc->origin[2]= desc->save_origin[2];
  
  GD_cmfree(&desc->transfer,GD_compute_bits(desc,desc->type));
}   /* GD_close_ds */

