#include <geodata.h>
#include <geodata_private.h>
#include <cm/paris.h>
#include <stdio.h>
#include <math.h>
#include "vector.h"

#define MY_PI GS_PI -0.000001

/* wrap(dest)
   
   wrap a float (23,8) value to the range (-GS_PI,GS_PI)
   
   CM_field_id_t dest    the value to be wrapped.
   
   precond: none
   postcond: dest is wrapped.

*/

wrap(dest)
     CM_field_id_t dest;
{
  
  CM_field_id_t old_context;
  
  old_context = CM_allocate_stack_field(1);
  CM_store_context(old_context);
  /* wrap [-pi,-3pi] */
  CM_f_lt_constant_1L(dest,-GS_PI,23,8);
  CM_logand_context_with_test();
  CM_f_add_constant_2_1L(dest,2*GS_PI,23,8);
  /* wrap [-3pi,-5pi] */
  CM_f_lt_constant_1L(dest,-GS_PI,23,8);
  CM_logand_context_with_test();
  CM_f_add_constant_2_1L(dest,2*GS_PI,23,8);
  CM_load_context(old_context);
  /* wrap [pi,3pi] */
  CM_f_gt_constant_1L(dest,GS_PI,23,8);
  CM_logand_context_with_test();
  CM_f_subtract_constant_2_1L(dest,2*GS_PI,23,8);
  /* wrap [3pi,5pi] */
  CM_f_gt_constant_1L(dest,GS_PI,23,8);
  CM_logand_context_with_test();
  CM_f_subtract_constant_2_1L(dest,2*GS_PI,23,8);
  CM_load_context(old_context);
  CM_deallocate_stack_through(old_context);
}



/* 
  GD_long_line_filter( dest, proj, line, type, success)

  set success to context which does not contain "long line" points or vectors

  In whole earth projections, points which were close to each other
  are sometimes projected to distant parts of the screen.  In the case
  of vectors or points which are represented by contours, this can cause
  "long lines", lines drawn across the screen from one side of the
  projection to the other.  
  
  Parameters:
       dest is pointer to transfer field

       proj is the prohjection used

       line is the center longitude for the projection.
       
       type is the data type

       success is the field which is to be set as a mask of "long lines"

  Precond: all values in dest are in the form  lat,lon,alt,val
       and in RADIAN, RADIAN, METER units.
  
  Logic: call appropriate filter.

  postcond: success is set to all non "long line" vectors or points

  Devin Jones
  
*/

void GD_long_line_filter( dest, proj, line, type, success)
CM_field_id_t dest;
int proj;
float line;
int type;
CM_field_id_t success;
{
  CM_store_context(success);
  if(proj==GS_MILLER || proj==GS_MOLLWEIDE ) {
    if(proj==GS_MILLER) line=0.0; /* sometimes miller params are garbage. dj*/
    if(type==GD_VECTOR) GD_long_line_vector_filter(dest,line,success);
    else GD_long_line_mesh_filter(dest,line,success);
  }
}



/******************************************************************
  GD_long_line_vector_filter( dest, line, success) 

  filter vectors for "long lines".

  Parameters:
       dest is pointer to transfer field

       line is the center longitude for the projection.

       success is context of non "long line" vectors.

  precond all values in dest are in the form  lat,lon,alt,val
       and in RADIAN, RADIAN, METER units.
  
  logic long lines are a subset of cross lines, lines which cross 
        either the zero longitude of the projection or the edge of the
	projection (at PI or -PI )
	Find the cross lines
	Find the long lines which cross pi or negative pi.
	set success to eliminate the long lines.

  postcond: success is set to all non "long line" vectors
  
  Devin Jones
  
*******************************************************************/

GD_long_line_vector_filter( dest, line, success)
CM_field_id_t dest;
float line;
CM_field_id_t success;
{
  
  CM_field_id_t tlon1,tlon2,tot_lon,M_lon,M_lat,abs_tlon1,abs_tlon2;
  CM_field_id_t cross_lines,long_lines,not_long_lines;
  CM_field_id_t old_context;

  old_context = CM_allocate_stack_field(1);
  cross_lines = CM_allocate_stack_field(1);
  long_lines = CM_allocate_stack_field(1);
  tlon1 = CM_allocate_stack_field(32);
  tlon2 = CM_allocate_stack_field(32);
  abs_tlon1 = CM_allocate_stack_field(32);
  abs_tlon2 = CM_allocate_stack_field(32);
  tot_lon = CM_allocate_stack_field(32);

  CM_store_context(old_context);
  CM_u_move_const_always_1L(long_lines,0,1);
  CM_u_move_const_always_1L(cross_lines,0,1);

  /* wrap lon one store in tlon1*/
  CM_f_add_constant_3_1L(tlon1,dest+32,-line,23,8);
  wrap(tlon1);

  /* wrap lon two store in tlon2 */
  CM_f_add_constant_3_1L(tlon2,dest+5*32,-line,23,8);
  wrap(tlon2);
  
  /* isolate vectors with longitudes of opposite sign */
  CM_logxor_3_1L (cross_lines,tlon1+31,tlon2+31,1);
  CM_load_context(cross_lines);
  /* seperate 'short lines' which cross the viewing line from true */
  /* 'long lines' which cross the edge */
  CM_f_abs_2_1L(abs_tlon1,tlon1,23,8);
  CM_f_abs_2_1L(abs_tlon2,tlon2,23,8);
  CM_f_add_3_1L(tot_lon,abs_tlon1,abs_tlon2,23,8);
  CM_f_ge_constant_1L(tot_lon,MY_PI,23,8);
  CM_logand_context_with_test();
  CM_store_context(long_lines);
  CM_load_context(old_context);
  CM_lognot_1_1L(long_lines,1);
  CM_logand_context(long_lines);
  CM_store_context(success);
  CM_load_context(old_context);
  CM_deallocate_stack_through(old_context);
}/* end of evil line */




/* 
  GD_long_line_mesh_filter( dest, line, success) 

  filter point for "long lines".

  Parameters:
       dest is pointer to transfer field

       line is the center longitude for the projection.

       success is context of non "long line" points

  precond all values in dest are in the form  lat,lon,alt,val
       and in RADIAN, RADIAN, METER units.
  
  logic long lines are a subset of cross lines, lines which cross 
        either the zero longitude of the projection or the edge of the
	projection (at PI or -PI )
	Find the cross lines
	Find the long lines which cross pi or negative pi.
	set success to eliminate the long lines.

  postcond: success is set to all non "long line" points
  
  Devin Jones
  
*/

GD_long_line_mesh_filter(dest, line, success)
CM_field_id_t dest;
float line;
CM_field_id_t success;
{
/* logic:   add parm[0] to lon
   if lon gt pi subtract 2pi
   if lon lt -pi add 2pi
   test for sign change from point1 to point2
   clip those that change */

  
  CM_field_id_t tlon1,tlon2,tot_lon,M_lon,M_lat,abs_tlon1,abs_tlon2;
  CM_field_id_t cross_lines,long_lines,not_long_lines;
  CM_field_id_t old_context;
  CM_field_id_t temp1;

  old_context = CM_allocate_stack_field(1);
  cross_lines = CM_allocate_stack_field(1);
  long_lines = CM_allocate_stack_field(1);
  tlon1 = CM_allocate_stack_field(32);  /* self longitude */
  tlon2 = CM_allocate_stack_field(32);  /* right neighbor longitude */
  abs_tlon1 = CM_allocate_stack_field(32);
  abs_tlon2 = CM_allocate_stack_field(32);
  tot_lon = CM_allocate_stack_field(32);
  temp1 = CM_allocate_stack_field(32);

  CM_store_context(old_context);
  CM_u_move_const_always_1L(long_lines,0,1);
  CM_u_move_const_always_1L(cross_lines,0,1);

  /* wrap lon one store in tlon1*/
  CM_f_add_constant_3_1L(tlon1,dest+32,-line,23,8);
  wrap(tlon1);

  /* wrap lon two store in tlon2 */
  CM_get_from_news_always_1L (tlon2,tlon1,1,CM_upward,32);
  wrap(tlon2);
  
  /* isolate vectors with longitudes of opposite sign */
  CM_logxor_3_1L (cross_lines,tlon1+31,tlon2+31,1);
  CM_load_context(cross_lines);
  /* seperate 'short lines' which cross the viewing line from true 'long lines'
     which cross the edge */
  CM_f_abs_2_1L(abs_tlon1,tlon1,23,8);
  CM_f_abs_2_1L(abs_tlon2,tlon2,23,8);
  CM_f_add_3_1L(tot_lon,abs_tlon1,abs_tlon2,23,8);
  CM_f_gt_constant_1L(tot_lon,MY_PI,23,8);
  CM_logand_context_with_test();
  CM_store_context(long_lines);
  CM_load_context(old_context);
  CM_lognot_1_1L(long_lines,1);
  CM_logand_context(long_lines);
  CM_store_context(success);
  CM_load_context(old_context);
  CM_deallocate_stack_through(old_context);
}/* end of evil line */

