
/*  File Header  --------------------------------
    File Name:   mollweide.c
    Author and Date:  Gary Meeker  6/26/90
    Description:  Performs mollweide map projection. Contains
                  both serial and parallel code descriptions for both
		  the forward and reverse transformations.
    End Header   --------------------------------  */


#include <cm/paris.h>
#include <math.h>

#define PI 3.14159265358979323846
#define MAXERR 0.001


/*  Function Header  ----------------------------
    Description:  Performs mollweide projection on front end
                  machine. 

    Parameters:
                data: Array of float containing (lat, lon)
		parameters: Array of float containing projection parameters.
		            The only parameter is the center meridian, lambda0.
    Preconditions:  Data is in latitude, longitude form and longitude ranges
                    from -PI to PI. 
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly.
    Logic: 
    End Header       ----------------------------  */

void GD_fe_mollweide(data, success, parameters)
     float *data;
     int *success;
     float *parameters;
{
  float theta, thetap, dthetap, lambda;

  *success = 1;
  lambda = data[1];
  if (parameters[0] < 0.0)
    {
      if (lambda > (parameters[0] + PI)) lambda -= 2*PI;
    }
  else if (lambda < (parameters[0] - PI)) lambda += 2*PI;
         
/*Solve iteratively for theta*/
  thetap = data[0];
  dthetap = PI;
  while (fabs(dthetap) > MAXERR)
    {
      dthetap = -(thetap + sin(thetap) - PI*sin(data[0]))/(1.0 + cos(thetap));
      thetap += dthetap;
    }
  theta = thetap/2;
  data[0] = (lambda - parameters[0])*cos(theta)/PI;
  data[1] = 0.5*sin(theta);
}

/*  Function Header  ----------------------------
    Description:  Performs inverse mollweide projection on front end
                  machine.
    Parameters:
                data: Array of float containing (x, y)
		parameters: Array of float containing projection parameters.
		            The only parameter is the center meridian, lambda0.
    Preconditions:  Data is in 2-D rectangular form.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly. Result is in (lat,lon) form.
    Logic:
    End Header       ----------------------------  */

void GD_fe_inverse_mollweide(data, success, parameters)
     float *data;
     int *success;
     float *parameters;
{
  double theta, dlambda;
  
  *success = 1;
  if (fabs(data[1]) > 0.5)
    {
      data[1] = 0.5;
      *success = 0;
    }
  else
    {
      theta = asin(2*data[1]);
      dlambda = PI*data[0]/cos(theta);
      data[1] = parameters[0] + dlambda;
      data[0] = asin((2*theta + sin(2*theta))/PI);
      if (data[1] < -PI) data[1] +=2*PI;
      else if (data[1] > PI) data[1] -=2*PI;
      if (fabs(dlambda) > PI) 
	{
	  data[1] = PI;
	  *success = 0;
	}
    }
}


/*  Function Header  ----------------------------
    Description:  Performs mollweide projection on connection
                  machine. Maps all data on the hemisphere opposite
		  the hemisphere tangent to the observation plane 
		  to the equator.
    Parameters:
                data: CM_field_id_t pointing to array containing (lat, lon)
		parameters: Array of float containing projection parameters.
		            The only parameter is the center meridian, lambda0.
    Preconditions:  Data is in angular (lat, lon) form, and data field is in
                    current vp_set.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly. Result is in rectangular form.
    Logic:
    End Header       ----------------------------  */


void GD_mollweide (data, success, parameters)
     CM_field_id_t  data;
     CM_field_id_t  success;
     float  *parameters;
{
  int loop;

  CM_field_id_t x = data;
  CM_field_id_t y =  CM_add_offset_to_field_id(data, 32);
  CM_field_id_t theta = CM_allocate_stack_field(32);
  CM_field_id_t thetap = CM_allocate_stack_field(32);
  CM_field_id_t dthetap = CM_allocate_stack_field(32);
  CM_field_id_t temp = CM_allocate_stack_field(32);
  CM_field_id_t saved_context = CM_allocate_stack_field(1);
  
  CM_u_move_constant_1L(success, 1, 1);
/*Offset lambda to keep in range*/
  CM_store_context(saved_context);
  if (parameters[0] < 0.0)
    {
      CM_f_gt_constant_1L(y, (parameters[0] + PI), 23, 8);
      CM_logand_context_with_test();
      CM_f_subtract_constant_2_1L(y, 2*PI, 23, 8);
      CM_load_context(saved_context);
    }
  else 
    {
      CM_f_lt_constant_1L(y, (parameters[0] - PI), 23, 8);
      CM_logand_context_with_test();
      CM_f_add_constant_2_1L(y, 2*PI, 23, 8);
      CM_load_context(saved_context);
    }

/*Solve iteratively for theta*/
  CM_f_move_1L(thetap, x, 23, 8);
/* added loop variable -dj */
  loop = 1000;
  while ( (CM_global_count_context() != 0) && loop--)
    {
      CM_f_cos_2_1L(temp, thetap, 23, 8);
      CM_f_add_constant_2_1L(temp, 1.0, 23, 8);
      CM_f_sin_2_1L(theta, x, 23, 8);
      CM_f_multiply_constant_2_1L(theta, PI, 23, 8);
      CM_f_sin_2_1L(dthetap, thetap, 23, 8);
      CM_f_subfrom_2_1L(dthetap, theta, 23, 8);
      CM_f_subtract_2_1L(dthetap, thetap, 23, 8);
      CM_f_divide_2_1L(dthetap, temp, 23, 8);
      CM_f_add_2_1L(thetap, dthetap, 23, 8);
      CM_f_abs_1_1L(dthetap, 23, 8);
      CM_f_gt_constant_1L(dthetap, MAXERR, 23, 8);
      CM_logand_context_with_test();
    }
/*Compute the values*/
  if(!loop){ printf("mollweide did not converge for: \n");
	     GD_transfer_statistics(2,data); }
  CM_invert_context();
  CM_logand_context(saved_context);
  CM_store_context(success);
  CM_f_divide_constant_3_1L(theta, thetap, 2.0, 23, 8);
  CM_f_cos_2_1L(temp, theta, 23, 8);
  CM_f_subtract_constant_3_1L(x, y, parameters[0], 23, 8);
  CM_f_multiply_2_1L(x, temp, 23, 8);
  CM_f_divide_constant_2_1L(x, PI, 23, 8);
  
  CM_f_sin_2_1L(y, theta, 23, 8);
  CM_f_multiply_constant_2_1L(y, 0.5, 23, 8);
  CM_deallocate_stack_through(theta);
}



/*  Function Header  ----------------------------
    Description:  Performs inverse mollweide projection on connection
                  machine.
    Parameters:
                data: CM_field_id_t pointing to array containing 
		      2-D rectangular coordinates.
		parameters: Array of float containing projection parameters.
		            The only parameter is the center meridian, lambda0.
    Preconditions:  Data is in rectangular (x, y) form, and data 
                    field is in current vp_set.
    Postconditions:  None.
    Return value:  Performs operation on memory 
                   elements so no value returned explicitly.
		   Result is in angular (lat,lon) form.
    Logic:
    End Header       ----------------------------  */

void GD_inverse_mollweide(data, success, parameters)
     CM_field_id_t  data;
     CM_field_id_t success;
     float  *parameters;
{
  CM_field_id_t lat = data;
  CM_field_id_t lon =  CM_add_offset_to_field_id(data, 32);
  CM_field_id_t theta = CM_allocate_stack_field(32);
  CM_field_id_t temp = CM_allocate_stack_field(32);
  CM_field_id_t saved_context = CM_allocate_stack_field(1);

/*Check for out of range y*/  
  CM_u_move_constant_1L(success, 1, 1);
  CM_store_context(saved_context);
  CM_f_abs_2_1L(temp, lon, 23, 8);
  CM_f_gt_constant_1L(temp, 0.5, 23, 8);
  CM_logand_context_with_test();
  CM_f_move_constant_1L(lon, 0.5, 23, 8);
  CM_u_move_constant_1L(success, 0, 1);
  CM_load_context(saved_context);

  CM_f_multiply_constant_3_1L(theta, lon, 2.0, 23, 8);
  CM_f_asin_1_1L(theta, 23, 8);
  
  CM_f_multiply_constant_3_1L(lon, lat, PI, 23, 8);
  CM_f_cos_2_1L(temp, theta, 23, 8);
  CM_f_divide_2_1L(lon, temp, 23, 8);
  CM_f_move_1L(temp, lon, 23, 8);
  CM_f_add_constant_2_1L(lon, parameters[0], 23, 8);
  
  CM_f_multiply_constant_2_1L(theta, 2.0, 23, 8);
  CM_f_sin_2_1L(lat, theta, 23, 8);
  CM_f_add_2_1L(lat, theta, 23, 8);
  CM_f_divide_constant_2_1L(lat, PI, 23, 8);
  CM_f_asin_1_1L(lat, 23, 8);

  CM_store_context(saved_context);
  CM_f_lt_constant_1L(lon, -PI, 23, 8);
  CM_logand_context_with_test();
  CM_f_add_constant_2_1L(lon, 2*PI, 23, 8);
  CM_load_context(saved_context);

  CM_f_gt_constant_1L(lon, PI, 23, 8);
  CM_logand_context_with_test();
  CM_f_subtract_constant_2_1L(lon, 2*PI, 23, 8);
  CM_load_context(saved_context);
  
/*Check for temp > PI as this signifies a bad x,y coordinate*/
  CM_f_abs_1_1L(temp, 23, 8);
  CM_f_gt_constant_1L(temp, PI, 23, 8);
  CM_logand_context_with_test();
  CM_f_move_constant_1L(lon, PI, 23, 8);
  CM_u_move_constant_1L(success, 0, 1);
  CM_load_context(saved_context);

  CM_deallocate_stack_through(theta);  
}


void GD_mollweide_border(dest,parameters)
     CM_field_id_t dest;
     float *parameters;
{
  CM_my_news_coordinate_1L (dest+96,0,32);
  CM_f_u_float_2_2L (dest+64,dest+96,32,23,8);
  CM_f_mult_const_add_const_1L (dest+64,dest+64,
				2.0*PI/65536.0,0.0,23,8);
  CM_f_cos_2_1L (dest,dest+64,23,8);
  CM_f_sin_2_1L (dest+32,dest+64,23,8);
  CM_f_multiply_constant_2_1L (dest+32,0.5,23,8);
  CM_get_from_news_always_1L (dest+128,dest,0,CM_upward,64);
  CM_f_move_constant_1L (dest+64,0.0,23,8);
  CM_f_move_constant_1L (dest+96,0.0,23,8);
  CM_f_move_constant_1L (dest+64+128,0.0,23,8);
  CM_f_move_constant_1L (dest+96+128,0.0,23,8);
}
