
/*  File Header  --------------------------------
    File Name:   polar_stereo.c
    Author and Date:  Gary Meeker  6/7/90
    Description:  Performs polar stereographic map projection. Contains
                  both serial and parallel code descriptions for both
		  the forward and reverse transformations.
    End Header   --------------------------------  */


#include <cm/paris.h>
#include <math.h>

#define PI 3.14159265358979323846


/*  Function Header  ----------------------------
    Description:  Performs polar sterographic projection on front end
                  machine. Maps all data on the hemisphere opposite
		  the hemisphere tangent to the observation plane 
		  to the equator.
    Parameters:
                data: Array of float containing (lat, lon)
		parameters: Array of float containing projection parameters.
		            In this case only one parameter, 1.0 for north
			    polar and -1.0 for south polar.
    Preconditions:  Data is in latitude, longitude form.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly.
    Logic:  First, set latitude a data point in opposite hemisphere
            to equal zero.  In this way, all of these points are mapped
	    to the equator.
    End Header       ----------------------------  */

void GD_fe_polar_stereo(data, success, parameters)
     float *data;
     int *success;
     float *parameters;
{
  float  r;

  *success = 1;
/*Data on opposite side of the earth*/
  if((data[0]*(-parameters[0])) > 0.0) 
    {
      data[0] = 0.0;
      *success = 0;
    }
  r = tan((PI/4) - (parameters[0]*data[0]/2));
  data[0] = r*sin(data[1]);
  data[1] = -parameters[0]*r*cos(data[1]);
}


/*  Function Header  ----------------------------
    Description:  Performs polar inverse sterographic projection on front end
                  machine.
    Parameters:
                data: Array of float containing (x, y)
		parameters: Array of float containing projection parameters.
		            In this case only one parameter, 1.0 for north
			    polar and -1.0 for south polar.
    Preconditions:  Data is in 2-D rectangular form.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly. Result is in (lat,lon) form.
    Logic:
    End Header       ----------------------------  */

void GD_fe_inverse_polar_stereo(data, success, parameters)
     float *data;
     int *success;
     float *parameters;
{
  float c, x, y, rho;
  
  *success = 1;
  x = data[0];
  y = data[1];
  rho = sqrt(x*x + y*y);
  if (rho > 1.0)
    {
      x = y = rho = 0.0;
      *success = 0;
    }
  c = 2*atan(rho);
  data[0] = asin(parameters[0]*cos(c));
  y = -parameters[0]*y;
  data[1] = atan2(x,y);
}


/*  Function Header  ----------------------------
    Description:  Performs polar sterographic projection on connection
                  machine. Maps all data on the hemisphere opposite
		  the hemisphere tangent to the observation plane 
		  to the equator.
    Parameters:
                data: CM_field_id_t pointing to array containing (lat, lon)
		parameters: Array of float containing projection parameters.
		            In this case only one parameter, 1.0 for north
			    polar and -1.0 for south polar.
    Preconditions:  Data is in angular (lat, lon) form, and data field is in
                    current vp_set.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly. Result is in rectangular form.
    Logic:
    End Header       ----------------------------  */


void GD_polar_stereo (data, success, parameters)
     CM_field_id_t  data;
     CM_field_id_t  success;
     float  *parameters;
{
  CM_field_id_t  x = data;
  CM_field_id_t  y = CM_add_offset_to_field_id (data,32);
  CM_field_id_t  hcomp = CM_allocate_stack_field(1);
  CM_field_id_t  r = CM_allocate_stack_field(32);
  CM_field_id_t  saved_context = CM_allocate_stack_field(1);
  CM_field_id_t  hemi = CM_allocate_stack_field(1);

  CM_u_move_constant_1L(success, 1, 1);
  if (parameters[0] == 1.0) CM_u_move_constant_1L(hcomp,1,1);
  else CM_u_move_constant_1L(hcomp,0,1);

/*Map data from opposite hemisphere to equator*/
  CM_lognot_2_1L(hemi, hcomp, 1);
  CM_logxor_2_1L(hemi, CM_add_offset_to_field_id(x,31), 1);
  CM_store_context(saved_context);
  CM_logand_context(hemi);
  CM_f_move_zero_1L(x, 23, 8);
  CM_u_move_zero_1L(success, 1);
  CM_load_context(saved_context);

  CM_f_move_1L(r, x, 23, 8);
  CM_f_multiply_constant_2_1L(r, 0.5, 23, 8);
  CM_logxor_2_1L(CM_add_offset_to_field_id(r, 31), hcomp, 1);
  CM_f_add_constant_2_1L(r, PI/4, 23,8);
  CM_f_tan_1_1L(r, 23, 8);
  CM_f_sin_2_1L(x, y, 23, 8);
  CM_f_multiply_2_1L(x, r, 23, 8);
  CM_f_cos_1_1L(y, 23, 8);
  CM_f_multiply_2_1L(y, r, 23, 8);
  CM_logxor_2_1L(CM_add_offset_to_field_id(y,31), hcomp, 1);
  CM_deallocate_stack_through(hcomp);
}



/*  Function Header  ----------------------------
    Description:  Performs polar inverse sterographic projection on connection
                  machine.
    Parameters:
                data: CM_field_id_t pointing to array containing 
		      2-D rectangular coordinates.
		parameters: Array of float containing projection parameters.
		            In this case only one parameter, 1.0 for north
			    polar and -1.0 for south polar.
    Preconditions:  Data is in rectangular (x, y) form, and data 
                    field is in current vp_set.
    Postconditions:  None.
    Return value:  Performs operation on memory 
                   elements so no value returned explicitly.
		   Result is in angular (lat,lon) form.
    Logic:
    End Header       ----------------------------  */

void GD_inverse_polar_stereo(data, success, parameters)
     CM_field_id_t  data;
     CM_field_id_t success;
     float  *parameters;
{
  CM_field_id_t  lat = data;
  CM_field_id_t  lon = CM_add_offset_to_field_id(data, 32);
  CM_field_id_t  hcomp = CM_allocate_stack_field(1);
  CM_field_id_t  c = CM_allocate_stack_field(32);
  CM_field_id_t  temp = CM_allocate_stack_field(32);
  CM_field_id_t  saved_context = CM_allocate_stack_field(1);
  
  CM_u_move_constant_1L(success, 1, 1);
  if (parameters[0] == 1.0) CM_u_move_constant_1L(hcomp,1,1);
  else CM_u_move_constant_1L(hcomp,0,1);
  CM_f_u_power_constant_3_1L(temp, lat, 2, 23, 8);
  CM_f_u_power_constant_3_1L(c, lon, 2, 23, 8);
  CM_f_add_2_1L(c, temp, 23, 8);
  CM_f_sqrt_1_1L(c, 23, 8);
  
/*Check for out of bounds points, map to (0,0)*/
  CM_store_context(saved_context);
  CM_f_gt_constant_1L(c, 1.0, 23, 8);
  CM_logand_context_with_test();
  CM_f_move_zero_1L(lat, 23, 8);
  CM_f_move_1L(lon, lat, 23, 8);
  CM_f_move_1L(c, lat, 23, 8);
  CM_u_move_zero_1L(success, 1);
  CM_load_context(saved_context);

  CM_f_atan_1_1L(c, 23, 8);
  CM_f_multiply_constant_2_1L(c, 2.0, 23, 8);
  CM_logxor_2_1L(CM_add_offset_to_field_id(lon, 31), hcomp, 1);
  CM_lognot_1_1L(hcomp, 1);
  CM_f_atan2_3_1L(lon, lat, lon, 23, 8);
  CM_f_cos_1_1L(c, 23, 8);
  CM_logxor_2_1L(CM_add_offset_to_field_id(c, 31), hcomp, 1);
  CM_f_asin_2_1L(lat, c, 23, 8);
  CM_deallocate_stack_through(hcomp);
}


void GD_polar_stereo_border(dest,parameters)
     CM_field_id_t dest;
     float *parameters;
{
  CM_my_news_coordinate_1L (dest+96,0,32);
  CM_f_u_float_2_2L (dest+64,dest+96,32,23,8);
  CM_f_mult_const_add_const_1L (dest+64,dest+64,
				2.0*PI/65536.0,0.0,23,8);
  CM_f_cos_2_1L (dest,dest+64,23,8);
  CM_f_sin_2_1L (dest+32,dest+64,23,8);

  CM_get_from_news_always_1L (dest+128,dest,0,CM_upward,64);
  CM_f_move_constant_1L (dest+64,0.0,23,8);
  CM_f_move_constant_1L (dest+96,0.0,23,8);
  CM_f_move_constant_1L (dest+64+128,0.0,23,8);
  CM_f_move_constant_1L (dest+96+128,0.0,23,8);
}
